unit UMathVec;

interface

uses dglOpenGL,Math,ODEImport,OpenGL;

type
  // tipo vetor
  CVector3f = record
    x,y,z : GLfloat;
  end;

  CVector4f = record
    x,y,z,w : GLfloat;
  end;

  CColor4f = record
    r,g,b,a : GLfloat;
  end;

  ODEObject = record
    Geom : PdxGeom;
    Body : PdxBody;
    Size : CVector3f;
  end;

  TODEArray = array of ODEObject;

  CBox = record
    Pos:CVector3f;
    Size:CVector3f;
  end;

  // tipo luz
  CLight = record
    m_bSpotLight:boolean;

    m_fPosition: CVector4f;
    m_fSpotDirection: CVector3f;

    m_fQuad_Attenuation: GLfloat;
    m_fLin_Attenuation: GLfloat;
    m_fConst_Attenuation: GLfloat;

    m_fAmbient:CColor4f;
    m_fDiffuse:CColor4f;
    m_fSpecular:CColor4f;
    m_fGlobal:CColor4f;
  end;

  CMaterial = record
    m_fAmbient:CColor4f;
    m_fDiffuse:CColor4f;
    m_fSpecular:CColor4f;
    m_fEmission:CColor4f;
    m_fShininess:GLfloat;
  end;

  CRotation = record
    x,y,z :double;
    Enabled:boolean;
  end;
  
function BackBufferMouseClick(mX,mY:integer):CVector3f;
procedure ReadGLMatrix();
procedure setTransform (pos : PdVector3; R : PdMatrix3);
function AxisAngletoMatrix(degree,x,y,z:GLfloat):TdMatrix3;
procedure setTranslate10;

function toVector3f(x,y,z:GLfloat):CVector3f;
procedure DrawAxis();
procedure DrawGrid(num_lines:integer);
procedure DrawPlane();
procedure DrawColorCube(render: Integer);
procedure DrawCube(render: Integer);

procedure ODEInitStoneHenge(var ODEObj:TODEArray; space : PdxSpace; world : PdxWorld);
procedure ODEDrawStoneHenge(ODEObj:TODEArray);

implementation

// Calcula a posio do mouse em 3d no backbuffer (plano far)
// isso porque o paramentro winZ do glUnproject  1.0
function BackBufferMouseClick(mX,mY:integer):CVector3f;
var
    // tipos OpenGL.pas
    //viewport: array [0..3] of GLint;
    //modelview: array [0..15] of GLdouble;
    //projection: array [0..15] of GLdouble;

    // tipos dglOpenGl.pas
    viewport: TVector4i;
    modelview: TGLMatrixd4;
    projection: TGLMatrixd4;
    
    dx,dy,dz,RealY: GLdouble;
begin      
  glGetDoublev(GL_MODELVIEW_MATRIX, @modelview);
	glGetDoublev(GL_PROJECTION_MATRIX, @projection);
	glGetIntegerv(GL_VIEWPORT, @viewport);

  // flip y up side down
	RealY := viewport[3] - mY -1;

  //http://pyopengl.sourceforge.net/documentation/manual/gluUnProject.3G.html
	// get 3D coordinates based on window coordinates
  gluUnProject(mX, RealY, 1.0, @modelview, @projection, @viewport, dx, dy ,dz); //  OpenGL.pas
  //	gluUnProject(mX, RealY, 1.0, modelview, projection, viewport, @dx, @dy ,@dz); //  dglOpenGl.pas

	result.x := dx;
	result.y := dy;
	result.z := dz;
end;

procedure ReadGLMatrix();
var  m: array[0..15] of GLfloat;
begin
  glGetFloatv(GL_MODELVIEW_MATRIX,@m);
end;

procedure setTransform (pos : PdVector3; R : PdMatrix3);
var
  m: array[0..15] of GLfloat;
begin
  m[0] := R[0];    m[1] := R[4];    m[2] := R[8];    m[3] := 0;
  m[4] := R[1];    m[5] := R[5];    m[6] := R[9];    m[7] := 0;
  m[8] := R[2];    m[9] := R[6];    m[10] := R[10];  m[11] := 0;
  m[12] := pos[0]; m[13] := pos[1]; m[14] := pos[2]; m[15] := 1;

  glMultMatrixf (@m);
end;

procedure setTranslate10();
var
  m: array[0..15] of GLfloat;
begin

  m[0] := 1.0;    m[1] := 0.0;  m[2] := 0.0;   m[3] := 0.0;
  m[4] := 0.0;    m[5] := 1.0;  m[6] := 0.0;   m[7] := 0.0;
  m[8] := 0.0;    m[9] := 0.0;  m[10] := 1.0;  m[11] := 0.0;
  m[12] := 10.0;  m[13] := 0.0; m[14] := 0.0;  m[15] := 1.0;

  glMultMatrixf (@m);
end;

function AxisAngletoMatrix(degree,x,y,z:GLfloat):TdMatrix3;
var R:TdMatrix3;
    s,c,t:GLfloat;
begin
  //http://www.euclideanspace.com/maths/geometry/rotations/conversions/angleToMatrix/
  //http://pyopengl.sourceforge.net/documentation/manual/glRotate.3G.html

  s := sin(DegToRad(degree));
  c := cos(DegToRad(degree));
  t := 1-c;
  R[0]:= t*x*x + c;   R[4]:= t*x*y - z*s; R[8]:= t*x*z + y*s;
  R[1]:= t*x*y + z*s; R[5]:= t*y*y + c;   R[9]:= t*y*z - x*s;
  R[2]:= t*x*z - y*s; R[6]:= t*y*z + x*s; R[10]:= t*z*z + c;

  AxisAngletoMatrix := R;
end;

function toVector3f(x,y,z:GLfloat):CVector3f;
var v:CVector3f;
begin
  v.x := x;
  v.y := y;
  v.z := z;
  toVector3f := v;
end;

procedure DrawAxis();
begin
  glLineWidth(3);
  glEnable(GL_LINE_SMOOTH);
  glBegin(GL_LINES);
    // red x axis
    glColor3f(1,0,0);
    glVertex3f(0,0,0);
    glVertex3f(1,0,0);

    // green y axis
    glColor3f(0,1,0);
    glVertex3f(0,0,0);
    glVertex3f(0,1,0);

    // blue z axis
    glColor3f(0,0,1);
    glVertex3f(0,0,0);
    glVertex3f(0,0,1);
  glEnd();
  glLineWidth(1);
  glDisable(GL_LINE_SMOOTH);
end;

procedure DrawGrid(num_lines:integer);
var i:integer;
begin

  glColor3f(0,0.5,0);

  glBegin(GL_LINES);

    glNormal3f(0,0,1);
	  // Desenha um grid no plano x/y
	  for i := -num_lines to num_lines do
	  begin
		  // horizontal lines (direcao X)
      glVertex3f(-num_lines, i, 0);
      glVertex3f(num_lines, i, 0);

		  // vertical lines (direcao Y)
      glVertex3f(i, -num_lines, 0);
      glVertex3f(i, num_lines, 0);
	  end;

  glEnd();
end;

procedure face(render: Integer);
var u,v,du,dv: real; n:integer;
begin
glPushMatrix;
  u:=-1/2;
  v:=-1/2;
  glTranslate(0.5,0,0);
  if render = GL_QUADS then
    begin
    n := 20;
    du:=1/n;
    dv:=1/n;
    glEnable(GL_NORMALIZE);
    glNormal3f(1,0,0);
    while u<(1/2-du) do
      begin
      v:=-1/2;
      while v<(1/2-dv) do
        begin
        glBegin(render);
          glVertex3f(0,u,v);
          glvertex3f(0,u,v+dv);
          glvertex3f(0,u+du,v+dv);
          glvertex3f(0,u+du,v);
        glEnd;
        v:=v+dv;
        end;
        u:=u+du;
      end;
    end
    else
    begin
        du:=1;
        dv:=1;
        glBegin(render);
          glVertex3f(0,u,v);
          glvertex3f(0,u,v+dv);
          glvertex3f(0,u+du,v+dv);
          glvertex3f(0,u+du,v);
        glEnd;
    end;
glPopMatrix;
end;

procedure DrawPlane();
begin
glPushMatrix;
  glRotate(-90,0,1,0);
  glTranslated(-0.5,0,0);
  face(GL_QUADS);
glPopMatrix;
end;


procedure DrawCube(render: Integer);
begin
glPushMatrix;
  face(render);
  glRotate(90,0,0,1); face(render);
  glRotate(90,0,0,1); face(render);
  glRotate(90,0,0,1); face(render);
  glRotate(-90,0,1,0); face(render);
  glRotate(-180,0,1,0); face(render);
glPopMatrix;
end;

procedure DrawColorCube(render: Integer);
begin
glPushMatrix;
  glColor3f(1,0,0); face(render);
  glColor3f(0,1,0); glRotate(90,0,0,1); face(render);
  glColor3f(0,1,1); glRotate(90,0,0,1); face(render);
  glColor3f(1,0,1); glRotate(90,0,0,1); face(render);
  glColor3f(0,0,1); glRotate(-90,0,1,0); face(render);
  glColor3f(1,1,0); glRotate(-180,0,1,0); face(render);
glPopMatrix;
end;

procedure ODEInitStoneHenge(var ODEObj:TODEArray; space : PdxSpace; world : PdxWorld);
var rot, drot, radius:double; c:integer;
    m:TdMass;
begin
  // variavel de rotacao em zero
  rot:=0;

  drot := 12;
  drot := DegToRad(drot);

  // raio em que as pedras esto dispostas
  radius := 16.5;

  c := 0;

  // desenha as pedras na circunferencia
  while rot < 2*PI do
  begin
    SetLength(ODEObj,c+1);

    // pedra vertical

    ODEObj[c].Size := toVector3f(1.0,2.0,4.0);

    // create the box's collision geometry
    ODEObj[c].Geom := dCreateBox (space, 1.0,2.0,4.0);

    // create the box's physical body
    ODEObj[c].Body := dBodyCreate(world);

    // attach the geometry to the body
    dGeomSetBody(ODEObj[c].Geom, ODEObj[c].Body);

    // translada para a posio na circunferencia
    dBodySetPosition(ODEObj[c].Body, radius*sin(rot), 2, radius*cos(rot));

    // rotaciona a pedra em torno do seu eixo
    // face maior com vetor normal apontando
    // para o centro da circunferencia
    dBodySetRotation(ODEObj[c].Body,AxisAngletoMatrix(RadToDeg(rot),0,0,1));

    // escala   1.0,4.0,2.0
    dMassSetBoxTotal(m,1, 1.0,2.0,4.0);
    dBodySetMass(ODEObj[c].Body,@m);

    c := c+1;

    // incrementa a rotacao
    rot := rot + drot;
  end;

end;

procedure ODEDrawStoneHenge(ODEObj:TODEArray);
var i:integer;
    rm:PdMatrix3; //4 linhas com 3 colunas (4x3)
    pos:PdVector3;
begin
  for  i:=0  to Length(ODEObj)-1 do
  begin
   pos := dBodyGetPosition(ODEObj[i].Body);
   rm := dBodyGetRotation(ODEObj[i].Body);

    // desenha o cubo
    glPushMatrix();
      setTransform(pos,rm);
      glScalef(ODEObj[i].Size.x, ODEObj[i].Size.y , ODEObj[i].Size.z);
      DrawCube(GL_QUADS);
    glPopMatrix();
  end;

end;

end.
