unit unitMainForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, JvHidControllerClass, ComCtrls, ExtCtrls, Math, MMSystem,
  Buttons, Menus, IniFiles, unitWiinRemote, ShellAPI,
  AppEvnts, DateUtils, UConfiguraGL, OpenGL, ObjetosGL, Grids, Registry,
  OBJLoader, UMathVec, ODEImport, DB, DBClient;

type
  TMainForm = class(TForm)
    HIDCtl: TJvHidDeviceController;
    DrawTimer: TTimer;
    MotionTimer: TTimer;
    RepeatTimer: TTimer;
    BatteryTimer: TTimer;
    MainMenu1: TMainMenu;
    Exit1: TMenuItem;
    Calibrate1: TMenuItem;
    Preference1: TMenuItem;
    Options1: TMenuItem;
    AbrirCena1: TMenuItem;
    File1: TMenuItem;
    N3: TMenuItem;
    CursorTimer: TTimer;
    StatusBar1: TStatusBar;
    Panel1: TPanel;
    TimerGL: TTimer;
    ClientDataSet1: TClientDataSet;
    OpenDialog1: TOpenDialog;
    ReacaoHaptica1: TMenuItem;
    procedure HIDCtlDeviceChange(Sender: TObject);
    procedure WiiData(HidDev: TJvHidDevice; ReportID: Byte; const Data: Pointer; Size: Word);
    procedure DrawTimerTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MotionTimerTimer(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RepeatTimerTimer(Sender: TObject);
    procedure CalibrateButtonClick(Sender: TObject);
    procedure ShowMenuClick(Sender: TObject);
    procedure TrayIcon1DblClick(Sender: TObject);
    procedure ExitMenuClick(Sender: TObject);
    procedure BatteryTimerTimer(Sender: TObject);

    procedure IRResetButtonClick(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure CursorTimerTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure TimerGLTimer(Sender: TObject);
    procedure CursorOnOff1Click(Sender: TObject);
    procedure AbrirCena1Click(Sender: TObject);
    procedure ReacaoHaptica1Click(Sender: TObject);

  private

    bmpWiimote: TBitmap;

    CurrentDevice: TJvHidDevice;
    GraphData: array [0..1023] of TAxisData;

    ButtonPushFlagTrigger: Boolean;
    prevTriggerTime: Cardinal;

    RumbleFlag: Boolean;
    LastRumble: Cardinal;

    prevEdged: Boolean;
    MotionTimerFlag: Boolean;

    procedure AnalyzePacket(ReportID: Integer; ReportData: PChar; Size: Word);
    procedure PushXYZ(Data: PChar);
    procedure DrawImage;
    procedure SetWiimoteMode(Mode:Byte);
    procedure SendToDevice(Buf: array of Byte; WriteLength: Integer);
    procedure LEDCtrl(LEDState: Integer);
    procedure AnalyzeButtons(Buttons: Cardinal);
    procedure RefreshButtonStatus;
    procedure ButtonEvent(Button, EventType: Integer; var ButtonState: TButtonState);
    procedure BatteryRequest;
    procedure PushKeys(keys: string);
    procedure RaiseEvent(ButtonAssign: TButtonAssign; Button: Integer; EventType: Integer;
      var ButtonState: TButtonState; RepeatFlag: Boolean = False);
    procedure CursorOff;
    procedure CursorOn;

    procedure AnalyzeIR(ReportData: PChar);
    procedure AnalyzeNunchuk(ReportData: PChar);
  public
    procedure IRSensorStart;
    procedure NunchukStart;
  end;

type
  TCor = record
    R,G,B,a: real;
  end;

  TTipoObj = (otEsfera, otCapsula, otCubo, otPlano, otTriMesh);

  TObjetoCena = record
    P, Pant, V, A, eixoRot: TVet3d;
    Q: TdVector4;
    D1,D2,D3,D4,size, angRot: real;
    Modelo: TModel;
    cor: TCor;
    noCursor, selecionado, ODEMove, UserMove: boolean;
    MostraRastroSempre, MostraRastroSolto, MostraVecV, MostraVecA, MostraEixoRot: boolean;
    tipo: TTipoObj;
    MoveDir: array [1..3] of Boolean;
    rastro: array of TVet3d;
  end;

  TObjetosCena = record
    Objetos: array of TObjetoCena;
  end;

  TTestes = record
    p, pAlvo: array of TVet3d;
    tIni, TFim: TDateTime;
    size: array of real;
  end;

  TMatriz3x3 = array[1..3,1..3] of real;
  TMatriz4x4 = array[1..4,1..4] of real;

const
  dLEDbarra = 0.20; //m
  HFOV = 41; //
  VFOV = 31; //
  CURSORNOOBJ = 1.5; // 100x% do lado do obj
  CURSORNOALVO = 0.8; // 100x% do lado do obj
  DATHRESHOLD = 3;

var
  MainForm: TMainForm;
  debug: Boolean = false;
  tempoA: TTime;
  c3d: TCursor3D;
  pontos: Tvet3d;
  vpontos: Tvet3d;
  rastro: array of Tvet3d;
  VPN: Tvet3d;
  VRP: Tvet3d = (50,50,25);
  VUP: Tvet3d = (0,0,1);
  ALVO: Tvet3d = (0,0,0);

  t: real;
  angCamGL: real = pi/4;
  hCamGL: real = 25;
  fator: integer = 100;
  Lw,Gw,Pw: Tvet3d;
  Base, BaseEuler: TMatriz3x3;
  alfa,beta,gama: real;
  objetosCena: TObjetosCena;
  pinca: boolean = false;
  soltou: boolean = false;
  testeAtual: integer = 1;
  QUANT_TESTES: integer = 0;
  TestesResult: array of TTestes;
  CoresObj: array of TCor;
  CorrigeMovimento: array [1..3] of real = (1,2,1);
  destro: boolean = true;

  m_ODEWorld : PdxWorld;
  m_ODESpace : PdxSpace;
  contactgroup : TdJointGroupID;

  FloorBox : PdxGeom;
  vFloorBox : CBox;

  Size: TdVector3;
  Vertices: array [0..65535-1] of TdVector3;
  Indices: array [0..65535-1] of integer;
  TriMeshData: PdxTriMeshData;
  vPosTriMesh : CBox;
  Geom: array of PdxGeom;
  Body: array of PdxBody;
  TipoDeColisao: integer;
  TemGravidade: boolean = true;
  Atrito: array of real;

  ODEObj: TODEArray;

  m_L0:CLight;
  m_L1:CLight;
  m_Mt1:CMaterial;

implementation

uses
  UArquivoCena;

{$R *.dfm}

procedure GetMinMax(var M : TModel);
var I : Integer;
begin
  // Find the min and max in each coordinate system
  with M do
  begin
    BoundingBox.min.X := 0;
    BoundingBox.min.Y := 0;
    BoundingBox.min.Z := 0;
    BoundingBox.med := BoundingBox.min;
    BoundingBox.max := BoundingBox.min;

    For I :=1 to Vertices do
    begin
      if Vertex[I].X < BoundingBox.min.X then BoundingBox.min.X := Vertex[I].X
      else if Vertex[I].X > BoundingBox.max.X then BoundingBox.max.X := Vertex[I].X;

      if Vertex[I].Y < BoundingBox.min.Y then BoundingBox.min.Y := Vertex[I].Y
      else if Vertex[I].Y > BoundingBox.max.Y then BoundingBox.max.Y := Vertex[I].Y;

      if Vertex[I].Z < BoundingBox.min.Z then BoundingBox.min.Z := Vertex[I].Z
      else if Vertex[I].Z > BoundingBox.max.Z then BoundingBox.max.Z := Vertex[I].Z;
    end;

  BoundingBox.med.X := (BoundingBox.max.X+BoundingBox.min.X)/2;
  BoundingBox.med.Y := (BoundingBox.max.Y+BoundingBox.min.Y)/2;
  BoundingBox.med.Z := (BoundingBox.max.Z+BoundingBox.min.Z)/2;

  if BoundingBox.max.X-BoundingBox.min.X > BoundingBox.max.Y-BoundingBox.min.Y
    then BoundingBox.maxdim := BoundingBox.max.X-BoundingBox.min.X
    else BoundingBox.maxdim := BoundingBox.max.Y-BoundingBox.min.Y;
  if BoundingBox.max.Z-BoundingBox.min.Z > BoundingBox.maxdim
    then BoundingBox.maxdim := BoundingBox.max.Z-BoundingBox.min.Z

  end;
end;

procedure nearCallback (data : pointer; o1, o2 : PdxGeom); cdecl;
const
  // Number of collision points to allow, 1 is unstable, 3 is ok, 4 is good for boxstacking.
  // Howver, this comes at a price, more collisions mean more constraints, and execution time
  // memory usage grows O(^3). (3 is 9 times as bad as 1)
  cCOL_MAX = 5;

  // ***********************
  // ver atrito - coliso
  // http://demura.net/english/ode12  -> Friction pyramid approximation
  // http://ode-wiki.org/wiki/index.php?title=HOWTO_simple_bouncing_sphere
  // http://educacao.uol.com.br/disciplinas/fisica/colisoes-2-tipos.htm
  // http://blog.gmane.org/gmane.comp.lib.ode/page=9
  // http://www.brasilescola.com/fisica/colisoes.htm
  //
  // ver tipo de contato -> TdContactType
  // http://ode-wiki.org/wiki/index.php?title=Manual:_Joint_Types_and_Functions
  // http://www.fisica.ufpb.br/~romero/pdf/10_colisoes.pdf
  // http://www.fisicaevestibular.com.br/Dinamica21.htm
  // ***********************
var
  i : integer;
  b1, b2 : PdxBody;
  numc : integer;
  contact : array[0..cCOL_MAX-1] of TdContact;
  c : TdJointID;
begin
  // Retrieve the bodies for the geoms. The geoms don't have to have bodies, they can be
  // planes for instance, plances don't have bodies. Any geom can be created without a body,
  // but it will be solid as a rock. (won't budge no matter what)
  b1 := dGeomGetBody(o1);
  b2 := dGeomGetBody(o2);

  // This test says that if both geoms have bodies, and they are connected with a joint,
  // then they should not collide.
  if (assigned(b1) and assigned(b2) and (dAreConnected(b1,b2)<>0)) then
    exit;

  // Here you can make other "early out" decision, if you decide that collision
  // should be ignored for these two geoms

  // Prepare the collision structure
  for i :=0 to cCOL_MAX-1 do
    begin
    contact[i].surface.mode := dContactBounce or dContactSoftCFM or dContactSoftERP;

    //contact[i].surface.mu := 500000;
    contact[i].surface.mu := atrito[i];
    contact[i].surface.mu2 := 0;

    contact[i].surface.soft_cfm := 0;      // borracha
    contact[i].surface.soft_erp := 0.5;    // ?? [0,2; 0,8]

    //contact[i].surface.bounce_vel := 0.1;
    contact[i].surface.bounce_vel := 0;

    case TipoDeColisao of // coeficiente de restituio [0; 1]
    0: contact[i].surface.bounce := 0.0; // Choque inelstico
    1: contact[i].surface.bounce := 0.5; // Choque parcialmente elstico
    2: contact[i].surface.bounce := 1.0; // Choque perfeitamente elstico
    end;
    
  end;
  // Call the more expensive collision method. numc will be set to the number of
  // actual collision returned. A box standing on a plane sureface needs three collisions
  // to be stable. A sphere only needs one.
  numc := dCollide (o1,o2,cCOL_MAX,contact[0].geom,sizeof(TdContact));

  // numc will be zero if the objects didn't actually collide.
  if (numc>0) then
  begin
    // If you'd like to add a sound to note the collision to the user, this is a good spot
    for i := 0 to numc-1 do
    begin
      // Add a contact joint that will prevent the two geoms from inter-penetrating
      c := dJointCreateContact(m_ODEWorld, contactgroup, @contact[i]);
      // Attach the joint to the two bodies. Remember than only one of them _must_ be a
      // body. One may be nil. But dJointAttach is prepared for that.
      dJointAttach (c,b1,b2);
    end;
  end;
end;

procedure InitODE();
begin
  m_ODEWorld := dWorldCreate();           // create world
  m_ODESpace := dHashSpaceCreate(nil);    // cria a forma de gerenciamento do espaco

  if TemGravidade                          // Set the world's global gravity vector.
    then dWorldSetGravity(m_ODEWorld,0,0,-9.810)
    else dWorldSetGravity(m_ODEWorld,0,0,0);
    
  dWorldSetCFM(m_ODEWorld,1e-5);          // Set the global CFM (constraint force mixing) value
//  dWorldSetERP(m_ODEWorld,0.5);

  vFloorBox.Size := toVector3f(100,100,0.5);
  vFloorBox.Pos := toVector3f(0,0,-0.25);
  // cria uma base
  FloorBox := dCreateBox(m_ODESpace, vFloorBox.Size.x, vFloorBox.Size.y , vFloorBox.Size.z);
  // posiciona o chao
  dGeomSetPosition(FloorBox, vFloorBox.Pos.x,vFloorBox.Pos.y,vFloorBox.Pos.z);

  contactgroup := dJointGroupCreate(0);
end;

procedure ShutDownODE();
begin
  dJointGroupDestroy (contactgroup);
  dSpaceDestroy (m_ODESpace);
  dWorldDestroy (m_ODEWorld);
end;

procedure UpdateODE();
const
  TIME_STEP = 0.01;
begin
  dSpaceCollide(m_ODESpace,nil,nearCallback);

  dWorldQuickStep(m_ODEWorld, TIME_STEP);
  //dWorldStep(m_ODEWorld, TIME_STEP);
  dJointGroupEmpty (contactgroup);
end;

procedure LimpaOBJ(Filename: string);
var
  arq1, arq2: TextFile;
  s, t: string;
begin
  AssignFile(arq1,Filename);
  Reset(arq1);

  AssignFile(arq2,'limpando OBJ.txt');
  Rewrite(arq2);

  while not(EOF(arq1)) do
    begin
    Readln(arq1,s);
    if pos('\',s) <> 0 then
      begin
      t := copy(s,1,Pos('\',s)-1);
      Readln(arq1,s);
      s := t + s;
      end;

    Writeln(arq2,s);
    end;

  CloseFile(arq1);
  CloseFile(arq2);

  DeleteFile(Filename);
  RenameFile('limpando OBJ.txt',Filename);
  DeleteFile('limpando OBJ.txt');
end;

procedure VerticesUnicosOBJ(var Modelo:TModel);
type
  Tdp = record
    de, para: integer;
  end;
var
  i,j,k: integer;
  troca: array of Tdp;
begin
  SetLength(troca,0);

  // monta tabela de-para com os vrtices repetidos
  for i := 1 to Modelo.Vertices do
    for j := i to Modelo.Vertices do
      if i <> j then
        if (Modelo.Vertex[i].X = Modelo.Vertex[j].X) and
           (Modelo.Vertex[i].Y = Modelo.Vertex[j].Y) and
           (Modelo.Vertex[i].Z = Modelo.Vertex[j].Z) then
          begin
          SetLength(troca, length(troca)+1);
          troca[length(troca)-1].de := j;
          troca[length(troca)-1].para := i;
          end;

  // procura chamadas recursivas  v1->v2->v3
  for i := 0 to Length(troca)-1 do
    for j := 1 to Length(troca)-1 do
    if i <> j then
      if troca[i].para = troca[j].de then
        troca[i].para := -1;

  // apaga chamadas recursivas v1->v2->v3 para v1->v3 e v2->v3
  i := 0;
  while i <= Length(troca)-1 do
    begin
    if troca[i].para = -1 then
      begin
      for j := i to Length(troca)-1 do
        troca[j] := troca[j+1];
      SetLength(troca,length(troca)-1);
      end;
    inc(i);
    end;

  // faz a troca nas faces do modelo, mas os vrtices no sero apagados da lista de vrtices
  for i := 1 to Modelo.Groups do
    for j := 1 to Modelo.Group[i].Faces do
      for k := 0 to Length(troca)-1 do
        if Modelo.Group[i].Face[j].vIndex[0] = troca[k].de then
          Modelo.Group[i].Face[j].vIndex[0] := troca[k].para;

end;

procedure EscalaOBJ(var Modelo: TModel; Sx, Sy, Sz: real);
var i: integer;
begin
  for i := 1 to Modelo.Vertices do
    begin
    Modelo.Vertex[i].X := Modelo.Vertex[i].X * Sx;
    Modelo.Vertex[i].Y := Modelo.Vertex[i].Y * Sy;
    Modelo.Vertex[i].Z := Modelo.Vertex[i].Z * Sz;
    end;
end;

procedure cubo(render: Integer);
  procedure face(render: Integer);
  var u,v,du,dv: real; n:integer;
  begin
  u:=-1/2;
  v:=-1/2;
  if render = GL_QUADS then
    begin
    n := 20;
    du:=1/n;
    dv:=1/n;
    glEnable(GL_NORMALIZE);
    glNormal3f(1,0,0);
    while u<(1/2-du) do
      begin
      v:=-1/2;
      while v<(1/2-dv) do
        begin
        glBegin(render);
          glVertex3f(1/2,u,v);
          glvertex3f(1/2,u,v+dv);
          glvertex3f(1/2,u+du,v+dv);
          glvertex3f(1/2,u+du,v);
        glEnd;
        v:=v+dv;
        end;
        u:=u+du;
      end;
    end
    else
    begin
        du:=1;
        dv:=1;
        glBegin(render);
          glVertex3f(1/2,u,v);
          glvertex3f(1/2,u,v+dv);
          glvertex3f(1/2,u+du,v+dv);
          glvertex3f(1/2,u+du,v);
        glEnd;
    end;
  end;

begin
face(render);
glRotate(90,0,0,1); face(render);
glRotate(90,0,0,1); face(render);
glRotate(90,0,0,1); face(render);
glRotate(-90,0,1,0); face(render);
glRotate(-180,0,1,0); face(render);
glLoadIdentity;
end;



procedure CalculaBase;
var m,xL,yB,xR,yT: real;
    c: integer;
begin
xL := min(Wiimote.Sensor.IR.dot1X, Wiimote.Sensor.IR.dot2X);
xR := max(Wiimote.Sensor.IR.dot1X, Wiimote.Sensor.IR.dot2X);
yB := min(Wiimote.Sensor.IR.dot1Y, Wiimote.Sensor.IR.dot2Y);
yT := max(Wiimote.Sensor.IR.dot1Y, Wiimote.Sensor.IR.dot2Y);

// versor L
Lw[1] := xR-xL;
Lw[2] := 0;
Lw[3] := yT-yB;
m := sqrt(sqr(Lw[1])+sqr(Lw[2])+sqr(Lw[3]));
for c := 1 to 3 do
  Lw[c] := Lw[c]/m;

// versor g
gw[1] := Wiimote.Sensor.Motion.X-127;
gw[2] := Wiimote.Sensor.Motion.Y-127;
gw[3] := Wiimote.Sensor.Motion.Z-127;
m := sqrt(sqr(gw[1])+sqr(gw[2])+sqr(gw[3]));
for c := 1 to 3 do
  gw[c] := gw[c]/m;

// versor p
pw[1] := gw[2]*Lw[3] - gw[3]*Lw[2];
pw[2] := gw[3]*Lw[1] - gw[1]*Lw[3];
pw[3] := gw[1]*Lw[2] - gw[2]*Lw[1];
m := sqrt(sqr(gw[1])+sqr(gw[2])+sqr(gw[3]));
for c := 1 to 3 do
  pw[c] := pw[c]/m;

base[1,1] := Lw[1];
base[1,2] := pw[1];
base[1,3] := gw[1];
base[2,1] := Lw[2];
base[2,2] := pw[2];
base[2,3] := gw[2];
base[3,1] := Lw[3];
base[3,2] := pw[3];
base[3,3] := gw[3];

end;

procedure AngEuler;
var gx,gy,gz: real;
begin
gx := Setting.Calibration.MotionSensor.X-127;
gy := Setting.Calibration.MotionSensor.Y-127;
gz := Setting.Calibration.MotionSensor.Z-127;

gx := gx/sqrt(sqr(gx)+sqr(gy)+sqr(gz));
gy := gy/sqrt(sqr(gx)+sqr(gy)+sqr(gz));
gz := gz/sqrt(sqr(gx)+sqr(gy)+sqr(gz));

// g dot eixos
alfa := ArcCos(base[1,1]*gx+base[1,2]*gy+base[1,3]*gz);
beta := ArcCos(base[2,1]*gx+base[2,2]*gy+base[2,3]*gz);
gama := ArcCos(base[3,1]*gx+base[3,2]*gy+base[3,3]*gz);
end;

procedure CalculaBaseEuler;
begin
//http://www2.dem.inpe.br/mcr/Inpe/CMC-202-4/pdf/aula07.pdf
BaseEuler[1,1] := +cos(alfa)*cos(beta)-sin(alfa)*cos(gama)*sin(beta);
BaseEuler[1,2] := +sin(alfa)*cos(beta)+cos(alfa)*cos(gama)*sin(beta);
BaseEuler[1,3] := +sin(gama)*sin(beta);
BaseEuler[2,1] := -cos(alfa)*cos(beta)-sin(alfa)*cos(gama)*cos(beta);
BaseEuler[2,2] := -sin(alfa)*sin(beta)+cos(alfa)*cos(gama)*cos(beta);
BaseEuler[2,3] := +sin(gama)*cos(beta);
BaseEuler[3,1] := +sin(alfa)*sin(gama);
BaseEuler[3,2] := -cos(alfa)*sin(gama);
BaseEuler[3,3] := +cos(gama);
end;

function matId(n: integer): TMatriz4x4;
var i,j: integer;
begin
for i := 1 to n do
  for j := 1 to n do
    Result[i,j] := 0;
for i := 1 to n do
  Result[i,i] := 1;
end;

function absVet(u:Tvet3d): real;
begin
Result := sqrt(sqr(u[1])+sqr(u[2])+sqr(u[3]));
end;

function somaVet(u, v:Tvet3d): TVet3d;
var i: integer;
begin
for i := 1 to 3 do
  Result[i] := u[i]+v[i];
end;

function difVet(u, v:Tvet3d): TVet3d;
var i: integer;
begin
for i := 1 to 3 do
  Result[i] := u[i]-v[i];
end;

function prodVet(u,v: Tvet3d): Tvet3d;
begin
Result[1] := u[2]*v[3] - u[3]*v[2];
Result[2] := u[3]*v[1] - u[1]*v[3];
Result[3] := u[1]*v[2] - u[2]*v[1];
end;

function prodEscVet(a: real; u: Tvet3d): Tvet3d;
var i: integer;
begin
for i := 1 to 3 do
  Result[i] := a*u[i];
end;

function normalizaVet(u: Tvet3d): Tvet3d;
var i: integer;
    m: real;
begin
m := absVet(u);
for i := 1 to 3 do
  Result[i] := u[i]/m;
end;

function matProd(m1,m2: TMatriz4x4): TMatriz4x4;
var i,j,k: integer;
begin
for i := 1 to 4 do
  for j := 1 to 4 do
    begin
    Result[i,j] := 0;
    for k := 1 to 4 do
      Result[i,j] := Result[i,j] + m1[i,k]*m2[k,j];
    end;
end;

function prodMatVet(m: TMatriz4x4; u: Tvet3d): Tvet3d;
var i,k: integer;
begin
for i := 1 to 4 do
  begin
  Result[i] := 0;
  for k := 1 to 4 do
    Result[i] := Result[i] + m[i,k]*u[k];
  end;
end;

function matTranslacao(v: TVet3d): TMatriz4x4;
var i: integer;
begin
Result := matId(4);
for i := 1 to 3 do
  Result[i,4] := v[i];
end;

function matRotacao(vpn, vup: Tvet3d): TMatriz4x4;
var rx,ry,rz: Tvet3d;
    i: integer;
begin
Result := matId(4);

rz := normalizaVet(vpn);
rx := prodVet(vup,rz);
rx := normalizaVet(rx);
ry := prodVet(rz,rx);

for i := 1 to 3 do
  begin
  Result[1,i] := rx[i];
  Result[2,i] := ry[i];
  Result[3,i] := rz[i];
  end;
end;

function matInv(a: TMatriz4x4): TMatriz4x4;
begin
// geogebra 4.2
Result[1,1] := (a[2,2]*a[3,3]*a[4,4] - a[2,2]*a[3,4]*a[4,3] - a[2,3]*a[3,2]*a[4,4] + a[2,3]*a[3,4]*a[4,2] + a[2,4]*a[3,2]*a[4,3] - a[2,4]*a[3,3]*a[4,2]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[1,2] := (-a[1,2]*a[3,3]*a[4,4] + a[1,2]*a[3,4]*a[4,3] + a[1,3]*a[3,2]*a[4,4] - a[1,3]*a[3,4]*a[4,2] - a[1,4]*a[3,2]*a[4,3] + a[1,4]*a[3,3]*a[4,2]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[1,3] := (a[1,2]*a[2,3]*a[4,4] - a[1,2]*a[2,4]*a[4,3] - a[1,3]*a[2,2]*a[4,4] + a[1,3]*a[2,4]*a[4,2] + a[1,4]*a[2,2]*a[4,3] - a[1,4]*a[2,3]*a[4,2]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[1,4] := (-a[1,2]*a[2,3]*a[3,4] + a[1,2]*a[2,4]*a[3,3] + a[1,3]*a[2,2]*a[3,4] - a[1,3]*a[2,4]*a[3,2] - a[1,4]*a[2,2]*a[3,3] + a[1,4]*a[2,3]*a[3,2]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[2,1] := (-a[2,1]*a[3,3]*a[4,4] + a[2,1]*a[3,4]*a[4,3] + a[2,3]*a[3,1]*a[4,4] - a[2,3]*a[3,4]*a[4,1] - a[2,4]*a[3,1]*a[4,3] + a[2,4]*a[3,3]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[2,2] := (a[1,1]*a[3,3]*a[4,4] - a[1,1]*a[3,4]*a[4,3] - a[1,3]*a[3,1]*a[4,4] + a[1,3]*a[3,4]*a[4,1] + a[1,4]*a[3,1]*a[4,3] - a[1,4]*a[3,3]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[2,3] := (-a[1,1]*a[2,3]*a[4,4] + a[1,1]*a[2,4]*a[4,3] + a[1,3]*a[2,1]*a[4,4] - a[1,3]*a[2,4]*a[4,1] - a[1,4]*a[2,1]*a[4,3] + a[1,4]*a[2,3]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[2,4] := (a[1,1]*a[2,3]*a[3,4] - a[1,1]*a[2,4]*a[3,3] - a[1,3]*a[2,1]*a[3,4] + a[1,3]*a[2,4]*a[3,1] + a[1,4]*a[2,1]*a[3,3] - a[1,4]*a[2,3]*a[3,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[3,1] := (a[2,1]*a[3,2]*a[4,4] - a[2,1]*a[3,4]*a[4,2] - a[2,2]*a[3,1]*a[4,4] + a[2,2]*a[3,4]*a[4,1] + a[2,4]*a[3,1]*a[4,2] - a[2,4]*a[3,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[3,2] := (-a[1,1]*a[3,2]*a[4,4] + a[1,1]*a[3,4]*a[4,2] + a[1,2]*a[3,1]*a[4,4] - a[1,2]*a[3,4]*a[4,1] - a[1,4]*a[3,1]*a[4,2] + a[1,4]*a[3,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[3,3] := (a[1,1]*a[2,2]*a[4,4] - a[1,1]*a[2,4]*a[4,2] - a[1,2]*a[2,1]*a[4,4] + a[1,2]*a[2,4]*a[4,1] + a[1,4]*a[2,1]*a[4,2] - a[1,4]*a[2,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[3,4] := (-a[1,1]*a[2,2]*a[3,4] + a[1,1]*a[2,4]*a[3,2] + a[1,2]*a[2,1]*a[3,4] - a[1,2]*a[2,4]*a[3,1] - a[1,4]*a[2,1]*a[3,2] + a[1,4]*a[2,2]*a[3,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[4,1] := (-a[2,1]*a[3,2]*a[4,3] + a[2,1]*a[3,3]*a[4,2] + a[2,2]*a[3,1]*a[4,3] - a[2,2]*a[3,3]*a[4,1] - a[2,3]*a[3,1]*a[4,2] + a[2,3]*a[3,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[4,2] := (a[1,1]*a[3,2]*a[4,3] - a[1,1]*a[3,3]*a[4,2] - a[1,2]*a[3,1]*a[4,3] + a[1,2]*a[3,3]*a[4,1] + a[1,3]*a[3,1]*a[4,2] - a[1,3]*a[3,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[4,3] := (-a[1,1]*a[2,2]*a[4,3] + a[1,1]*a[2,3]*a[4,2] + a[1,2]*a[2,1]*a[4,3] - a[1,2]*a[2,3]*a[4,1] - a[1,3]*a[2,1]*a[4,2] + a[1,3]*a[2,2]*a[4,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
Result[4,4] := (a[1,1]*a[2,2]*a[3,3] - a[1,1]*a[2,3]*a[3,2] - a[1,2]*a[2,1]*a[3,3] + a[1,2]*a[2,3]*a[3,1] + a[1,3]*a[2,1]*a[3,2] - a[1,3]*a[2,2]*a[3,1]) / (a[1,1]*a[2,2]*a[3,3]*a[4,4] - a[1,1]*a[2,2]*a[3,4]*a[4,3] - a[1,1]*a[2,3]*a[3,2]*a[4,4] + a[1,1]*a[2,3]*a[3,4]*a[4,2] + a[1,1]*a[2,4]*a[3,2]*a[4,3] - a[1,1]*a[2,4]*a[3,3]*a[4,2] - a[1,2]*a[2,1]*a[3,3]*a[4,4] + a[1,2]*a[2,1]*a[3,4]*a[4,3] + a[1,2]*a[2,3]*a[3,1]*a[4,4] - a[1,2]*a[2,3]*a[3,4]*a[4,1] - a[1,2]*a[2,4]*a[3,1]*a[4,3] + a[1,2]*a[2,4]*a[3,3]*a[4,1] + a[1,3]*a[2,1]*a[3,2]*a[4,4] - a[1,3]*a[2,1]*a[3,4]*a[4,2] - a[1,3]*a[2,2]*a[3,1]*a[4,4] + a[1,3]*a[2,2]*a[3,4]*a[4,1] + a[1,3]*a[2,4]*a[3,1]*a[4,2] - a[1,3]*a[2,4]*a[3,2]*a[4,1] - a[1,4]*a[2,1]*a[3,2]*a[4,3] + a[1,4]*a[2,1]*a[3,3]*a[4,2] + a[1,4]*a[2,2]*a[3,1]*a[4,3] - a[1,4]*a[2,2]*a[3,3]*a[4,1] - a[1,4]*a[2,3]*a[3,1]*a[4,2] + a[1,4]*a[2,3]*a[3,2]*a[4,1]);
end;

function matRotNoEixo(ang: real; axis: TVet3d): TMatriz4x4;
var leng, a, b, c, aa, bb, cc, sine, cosine, omcos: real;
    m: TMatriz4x4;
begin
  leng := sqrt(sqr(axis[1])+sqr(axis[2])+sqr(axis[3]));
  a := axis[1] / leng;
  b := axis[2] / leng;
  c := axis[3] / leng;
  aa := a * a;
  bb := b * b;
  cc := c * c;
  sine := sin(DegToRad(ang));
  cosine := cos(DegToRad(ang));
  omcos := 1 - cosine;

  m[1,1] := aa + (1 - aa) * cosine;
  m[1,2] := a * b * omcos - c * sine;
  m[1,3] := a * c * omcos + b * sine;
  m[1,4] := 0;

  m[2,1] := a * b * omcos + c * sine;
  m[2,2] := bb + (1 - bb) * cosine;
  m[2,3] := b * c * omcos - a * sine;
  m[2,4] := 0;

  m[3,1] := a * c * omcos - b * sine;
  m[3,2] := b * c * omcos + a * sine;
  m[3,3] := cc + (1 - cc) * cosine;
  m[3,4] := 0;

  m[4,1] := 0;
  m[4,2] := 0;
  m[4,2] := 0;
  m[4,3] := 1;

  Result := m;
end;

// Dolly da Cmera MTK (movimento pelo eixo da profundidade)
procedure MTK_Dolly(fw: real);
var v: TVet3d;
begin
v := normalizaVet(difVet(ALVO,VRP));
v := prodEscVet(fw,v);
VRP := somaVet(VRP,v);
end;

// Orbit da Cmera MTK
procedure MTK_Orbit(right, up: real);
var vright, vaux: TVet3d;
    m1, m2: TMatriz4x4;
    i: integer;
begin
vright := normalizaVet(prodVet(difVet(ALVO,VRP),VUP));
vup := normalizaVet(prodVet(vright,difVet(ALVO,VRP)));

m1 := matTranslacao(ALVO);
m1 := matProd(m1, matRotNoEixo({-}right,vup));
m1 := matProd(m1, matRotNoEixo(up,vright));
for i := 1 to 3 do
  vaux[i] := -Alvo[i];
m1 := matProd(m1, matTranslacao(vaux));

for i := 1 to 3 do
  vaux[i] := -VRP[i];
m2 := matTranslacao(vaux);
m2 := matProd(m2, m1);

VRP := prodMatVet(m1,VRP);
m2 := matProd(m2, matTranslacao(VRP));

// garante que up_  perpendicular ao vpn (aim-pos)
vright := normalizaVet(prodVet(difVet(Alvo,VRP),Vup));
Vup := normalizaVet(prodVet(vright,difVet(Alvo,VRP)));

end;

procedure CorrigeCanhoto(i:integer);
var aux: real;
begin
  // se canhoto, troca os eixos
  if not destro then
    begin
      aux := objetosCena.Objetos[i].p[1];
      objetosCena.Objetos[i].p[1] := objetosCena.Objetos[i].p[2];
      objetosCena.Objetos[i].p[2] := aux;
    end;
end;

// sinaliza que cursor est sobre um objeto
// alterando o campo "noCursor"
procedure CursorNoObjeto(p: TVet3d);
var i,j: integer;
    s: real;
begin
for i := 0 to Length(objetosCena.Objetos)-1 do
  begin
  if objetosCena.Objetos[i].UserMove then
    begin
    s := 0;
    for j := 1 to 3 do
      s := s + sqr(objetosCena.Objetos[i].p[j] - p[j]);
      objetosCena.Objetos[i].noCursor := sqrt(s) <= (CURSORNOOBJ*objetosCena.Objetos[i].size);
    end;
  end;
end;

// sinaliza que cursor est movimentando um objeto
procedure ObjetoSelecionado(p: TVet3d; pinca, rumble: boolean);
var i,j: integer;
    jaPegou: boolean;
    v: real;
    vec: TVet3d;
begin
  // desliga o rumble
  MainForm.LEDCtrl($90);
  MainForm.RumbleFlag := False;

  for i := 0 to Length(objetosCena.Objetos)-1 do
    begin
    if pinca then
      begin
      // garante que s segura um objeto da cena
      jaPegou := false;
      for j := 0 to Length(objetosCena.Objetos)-1 do
        if j <> i then
          jaPegou := jaPegou or objetosCena.Objetos[j].selecionado;

      if (not jaPegou) and (objetosCena.Objetos[i].noCursor) then
        begin
        objetosCena.Objetos[i].selecionado := true;

        // move obj com o wiimote
        for j := 1 to 3 do
          // com a restrio de movimento definido no arquivo
          if objetosCena.Objetos[i].MoveDir[j] then
            objetosCena.Objetos[i].p[j] := p[j];

        dBodySetPosition(Body[i],objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],objetosCena.Objetos[i].p[3]);


        // liga o rumble quando est segurando um objeto
        if rumble then
          begin
          MainForm.LEDCtrl($91);
          MainForm.LastRumble := timeGetTime;
          MainForm.RumbleFlag := True;
          end;
        end;
      end
      else
        begin

        if soltou and objetosCena.Objetos[i].selecionado then
          begin
            dBodySetForce(Body[i],0,0,0);
            vec := normalizaVet(difVet(objetosCena.Objetos[i].p,objetosCena.Objetos[i].pant));
            v := absVet(objetosCena.Objetos[i].v);
            vec := prodEscVet(v, vec);
          end;
        objetosCena.Objetos[i].selecionado := false;
        end;

    end;
end;

// desenha objetos na cena
procedure DesenhaObjetos(linhavertical: boolean);
const
 slices = 10;
 loops = 10;
 scaleVec = 1.2;
var
  i, n: integer;
  u, v, du, dv, tam: real;
  rm:PdMatrix3; //4 linhas com 3 colunas (4x3)
  pos:PdVector3;
  Q: GLUquadricObj;
begin
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  q := gluNewQuadric;

  for i := 0 to Length(objetosCena.Objetos)-1 do
    begin
    // definio da cor do objeto
    if objetosCena.Objetos[i].selecionado then
      begin
      glColor3f(1,0.5,0.5);
      end
      else
      if objetosCena.Objetos[i].noCursor
        then glColor3f(0.5,0.5,1)
        else
        begin
        if objetosCena.Objetos[i].cor.a = 0 then
          glColor3f(objetosCena.Objetos[i].cor.R/255,
                    objetosCena.Objetos[i].cor.G/255,
                    objetosCena.Objetos[i].cor.B/255)
        else
          begin
          glColor4f(objetosCena.Objetos[i].cor.R/255,
                    objetosCena.Objetos[i].cor.G/255,
                    objetosCena.Objetos[i].cor.B/255,
                    objetosCena.Objetos[i].cor.A/255);
          end;
        end;

    // desenha objeto real ou s um cubo para teste
    if debug then
      begin
      // desenha cubo
      glPushMatrix;
        glTranslate(objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],objetosCena.Objetos[i].p[3]);
        glscale(objetosCena.Objetos[i].size,objetosCena.Objetos[i].size,objetosCena.Objetos[i].size);
        cubo(GL_QUADS);
      glPopMatrix;
      end
      else
      begin
      pos := dBodyGetPosition(Body[i]);
      rm := dBodyGetRotation(Body[i]);
      glPushMatrix();
        setTransform(pos,rm);
        case objetosCena.Objetos[i].tipo of
        otEsfera: begin
          gluSphere(Q,objetosCena.Objetos[i].D1,slices,loops);
          end;
        otCapsula: begin
          glTranslatef(0,0,objetosCena.Objetos[i].D2 / 2);
          gluSphere(q,objetosCena.Objetos[i].D1,slices,loops);
          glTranslatef(0,0,-objetosCena.Objetos[i].D2);
          gluCylinder(q,objetosCena.Objetos[i].D1,objetosCena.Objetos[i].D1,objetosCena.Objetos[i].D2,slices,loops);
          gluSphere(q,objetosCena.Objetos[i].D1,slices,loops);
          end;
        otCubo: begin
        glRotate(radtodeg(objetosCena.Objetos[i].angRot),
                 objetosCena.Objetos[i].eixoRot[1],
                 objetosCena.Objetos[i].eixoRot[2],
                 objetosCena.Objetos[i].eixoRot[3]);
        glScaled(objetosCena.Objetos[i].D1,objetosCena.Objetos[i].D2,objetosCena.Objetos[i].D3);
          cubo(GL_QUADS);
          end;
        otPlano: begin
          n := 10;
          du:=objetosCena.Objetos[i].Modelo.BoundingBox.med.X/n;
          dv:=objetosCena.Objetos[i].Modelo.BoundingBox.med.Y/n;
          glEnable(GL_NORMALIZE);
          glNormal3f(objetosCena.Objetos[i].D1,objetosCena.Objetos[i].D2,objetosCena.Objetos[i].D3);
          u:=-1/2*objetosCena.Objetos[i].Modelo.BoundingBox.med.X;
          while u<(1/2*objetosCena.Objetos[i].Modelo.BoundingBox.med.X-du) do
            begin
            v:=-1/2*objetosCena.Objetos[i].Modelo.BoundingBox.med.Y;
            while v<(1/2*objetosCena.Objetos[i].Modelo.BoundingBox.med.Y-dv) do
              begin
              glColor3f(objetosCena.Objetos[i].cor.R/255,
                        objetosCena.Objetos[i].cor.G/255,
                        objetosCena.Objetos[i].cor.B/255);
              glBegin(GL_LINE_LOOP);
                glVertex3f(u,v,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*u+objetosCena.Objetos[i].D2*v)/objetosCena.Objetos[i].D3);
                glvertex3f(u,v+dv,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*u+objetosCena.Objetos[i].D2*(v+dv))/objetosCena.Objetos[i].D3);
                glvertex3f(u+du,v+dv,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*(u+du)+objetosCena.Objetos[i].D2*(v+dv))/objetosCena.Objetos[i].D3);
                glvertex3f(u+du,v,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*(u+du)+objetosCena.Objetos[i].D2*v)/objetosCena.Objetos[i].D3);
              glEnd;

              glColor4f(objetosCena.Objetos[i].cor.R/255,
                        objetosCena.Objetos[i].cor.G/255,
                        objetosCena.Objetos[i].cor.B/255,
                        0.5);
              glBegin(GL_QUADS);
                glVertex3f(u,v,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*u+objetosCena.Objetos[i].D2*v)/objetosCena.Objetos[i].D3);
                glvertex3f(u,v+dv,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*u+objetosCena.Objetos[i].D2*(v+dv))/objetosCena.Objetos[i].D3);
                glvertex3f(u+du,v+dv,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*(u+du)+objetosCena.Objetos[i].D2*(v+dv))/objetosCena.Objetos[i].D3);
                glvertex3f(u+du,v,-(objetosCena.Objetos[i].D4+objetosCena.Objetos[i].D1*(u+du)+objetosCena.Objetos[i].D2*v)/objetosCena.Objetos[i].D3);
              glEnd;

              v:=v+dv;
              end;
            u:=u+du;
            end;
          end;
        otTriMesh: begin
            glDisable(GL_COLOR_MATERIAL);
            DrawModel(objetosCena.Objetos[i].Modelo);
          end;
        end;

      glPopMatrix();
      end;

    glEnable(GL_COLOR_MATERIAL);
    end;

  // ******************************************************************
  //   desenhar objetos grficos auxiliares (linhas de guia, vetores)
  // ******************************************************************

  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_LIGHTING);

  for i := 0 to Length(objetosCena.Objetos)-1 do
    begin
    // desenha linha da altura
    if linhavertical then
      begin
        if objetosCena.Objetos[i].p[3] > 0
        then glColor3f(0,0,1)
        else glColor3f(1,0,0);
      glBegin(GL_LINES);
        glVertex3f(objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],objetosCena.Objetos[i].p[3]);
        glVertex3f(objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],0);
      glEnd;
      glBegin(GL_POINTS);
        glVertex3f(objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],0);
      glEnd;
    end;

    // desenha o vetor velocidade
    if objetosCena.Objetos[i].MostraVecV then
      begin
      glLineWidth(5);
      glPushMatrix;
        glTranslate(objetosCena.Objetos[i].p[1], objetosCena.Objetos[i].p[2], objetosCena.Objetos[i].p[3]);
        //glScale(objetosCena.Objetos[i].size,objetosCena.Objetos[i].size,objetosCena.Objetos[i].size);
        glscale(scaleVec,scaleVec,scaleVec);
        glBegin(GL_LINES);
          glColor3f(0,1,1);
          glVertex3f(0,0,0); glVertex3f(objetosCena.Objetos[i].v[1],0,0);
          glColor3f(1,0,1);
          glVertex3f(0,0,0); glVertex3f(0,objetosCena.Objetos[i].v[2],0);
          glColor3f(1,1,0);
          glVertex3f(0,0,0); glVertex3f(0,0,objetosCena.Objetos[i].v[3]);
          glColor3f(1,1,1);
          glVertex3f(0,0,0); glVertex3f(objetosCena.Objetos[i].v[1],objetosCena.Objetos[i].v[2],objetosCena.Objetos[i].v[3]);
        glEnd;
      glPopMatrix;
      glLineWidth(1);
      end;

    // desenha o vetor acelerao
    if objetosCena.Objetos[i].MostraVecA then
      begin
      glLineWidth(5);
      glPushMatrix;
        glTranslate(objetosCena.Objetos[i].p[1], objetosCena.Objetos[i].p[2], objetosCena.Objetos[i].p[3]);
        //glScale(objetosCena.Objetos[i].size,objetosCena.Objetos[i].size,objetosCena.Objetos[i].size);
        glscale(scaleVec,scaleVec,scaleVec);
        glBegin(GL_LINES);
          glColor3f(0,1,1);
          glVertex3f(0,0,0); glVertex3f(objetosCena.Objetos[i].a[1],0,0);
          glColor3f(1,0,1);
          glVertex3f(0,0,0); glVertex3f(0,objetosCena.Objetos[i].a[2],0);
          glColor3f(1,1,0);
          glVertex3f(0,0,0); glVertex3f(0,0,objetosCena.Objetos[i].a[3]);
          glColor3f(1,1,1);
          glVertex3f(0,0,0); glVertex3f(objetosCena.Objetos[i].a[1],objetosCena.Objetos[i].a[2],objetosCena.Objetos[i].a[3]);
        glEnd;
      glPopMatrix;
      glLineWidth(1);
      end;

    // eixo de rotao
    if  objetosCena.Objetos[i].MostraEixoRot then
      begin
      glLineWidth(5);
      glPushMatrix;
        glTranslate(objetosCena.Objetos[i].p[1], objetosCena.Objetos[i].p[2], objetosCena.Objetos[i].p[3]);
        //glScale(objetosCena.Objetos[i].size,objetosCena.Objetos[i].size,objetosCena.Objetos[i].size);
        //glscale(scaleVec,scaleVec,scaleVec);
        glscale(5,5,5);
        glBegin(GL_LINES);
          glColor3f(1,0.5,0.5);
          tam := sqrt(sqr(objetosCena.Objetos[i].q[1]) +
                      sqr(objetosCena.Objetos[i].q[2]) +
                      sqr(objetosCena.Objetos[i].q[3]));
          if tam <> 0 then
            begin
            glVertex3f(-abs(objetosCena.Objetos[i].q[1]/tam), -abs(objetosCena.Objetos[i].q[2]/tam), -abs(objetosCena.Objetos[i].q[3]/tam));
            glVertex3f(+abs(objetosCena.Objetos[i].q[1]/tam), +abs(objetosCena.Objetos[i].q[2]/tam), +abs(objetosCena.Objetos[i].q[3]/tam));
            end;
        glEnd;
      glPopMatrix;
      end;

    // anel de rotao
    {if  objetosCena.Objetos[i].MostraAnelRot then
      begin
      glLineWidth(5);
      glPushMatrix;
        pos := dBodyGetPosition(Body[i]);
        rm := dBodyGetRotation(Body[i]);
        setTransform(pos,rm);
        glRotate(90,1,0,0);
        //glScale(objetosCena.Objetos[i].size,objetosCena.Objetos[i].size,objetosCena.Objetos[i].size);
        //glscale(scaleVec,scaleVec,scaleVec);
        glscale(5,5,5);
        glBegin(GL_LINES);
          for arc := 0 to 17 do
            glVertex3f(1*cos(2*pi*arc/17), 1*sin(2*pi*arc/17), 0);
        glEnd;
      glPopMatrix;
      end;
      }
    glLineWidth(1);

    end;

  glDisable(GL_BLEND);

end;


procedure TMainForm.FormCreate(Sender: TObject);
var
  i: integer;
  ka, kd, ks, ke: array [1..4] of GLfloat;
begin
  Randomize;
  CriaBancoDeDados(ClientDataSet1);
  ConfiguraOpenGL(Panel1.Handle);
  LoadConfig;
  InitODE;

  MainForm.DoubleBuffered := True;

  bmpWiimote := TBitmap.Create;

  for i := 1 to 3 do
    pontos[i]:=0;

  t:=0;
  if Setting.MinimizeAtStartup then
    MainForm.WindowState := wsMinimized;

  SetLength(CoresObj,3);
  CoresObj[0].R := 1;
  CoresObj[0].G := 0;
  CoresObj[0].B := 1;
  CoresObj[1].R := 0;
  CoresObj[1].G := 1;
  CoresObj[1].B := 1;
  CoresObj[2].R := 1;
  CoresObj[2].G := 1;
  CoresObj[2].B := 0;

  // update light pos
  m_L0.m_fPosition.x := 5;
  m_L0.m_fPosition.y := -5;
  m_L0.m_fPosition.z := 3*abs(cos(0*t));
  m_L0.m_fPosition.w := 1;

  // update light pos
  m_L1.m_fPosition.x := -5;
  m_L1.m_fPosition.y := 5;
  m_L1.m_fPosition.z := 5;
  m_L1.m_fPosition.w := 0;

  // update spot light dir
  m_L1.m_fSpotDirection.x := 0;
  m_L1.m_fSpotDirection.y := -1;
  m_L1.m_fSpotDirection.z := -1;

  // light attenuation
  m_L0.m_fQuad_Attenuation  := 0.05;
  m_L0.m_fLin_Attenuation   := 0;
  m_L0.m_fConst_Attenuation := 0;

  m_L1.m_fQuad_Attenuation  := 0.05;
  m_L1.m_fLin_Attenuation   := 0;
  m_L1.m_fConst_Attenuation := 0;  

  m_L0.m_bSpotLight := false;

  m_L0.m_fAmbient.r := 0.1;
  m_L0.m_fAmbient.g := 0.1;
  m_L0.m_fAmbient.b := 0.1;
  m_L0.m_fAmbient.a := 1.0;

  m_L0.m_fDiffuse.r := 1.0;
  m_L0.m_fDiffuse.g := 1.0;
  m_L0.m_fDiffuse.b := 1.0;
  m_L0.m_fDiffuse.a := 1.0;

  m_L0.m_fSpecular.r := 0.2;
  m_L0.m_fSpecular.g := 0.2;
  m_L0.m_fSpecular.b := 0.2;
  m_L0.m_fSpecular.a := 1.0;

  m_L1.m_bSpotLight := true;

  m_L1.m_fAmbient.r := 0.1;
  m_L1.m_fAmbient.g := 0.1;
  m_L1.m_fAmbient.b := 0.1;
  m_L1.m_fAmbient.a := 1.0;

  m_L1.m_fDiffuse.r := 1.0;
  m_L1.m_fDiffuse.g := 1.0;
  m_L1.m_fDiffuse.b := 1.0;
  m_L1.m_fDiffuse.a := 1.0;

  m_L1.m_fSpecular.r := 0.2;
  m_L1.m_fSpecular.g := 0.2;
  m_L1.m_fSpecular.b := 0.2;
  m_L1.m_fSpecular.a := 1.0;

  // light position
  glLightfv(GL_LIGHT0, GL_POSITION, @m_L0.m_fPosition );
  glLightfv(GL_LIGHT1, GL_POSITION, @m_L1.m_fPosition );

  // spot direction
  if(m_L0.m_bSpotLight) then
    glLightfv( GL_LIGHT0, GL_SPOT_DIRECTION, @m_L0.m_fSpotDirection);
  if(m_L1.m_bSpotLight) then
    glLightfv( GL_LIGHT1, GL_SPOT_DIRECTION, @m_L1.m_fSpotDirection);

  // light components
  glLightfv(GL_LIGHT0,GL_AMBIENT,  @m_L0.m_fAmbient);
  glLightfv(GL_LIGHT0,GL_DIFFUSE,  @m_L0.m_fDiffuse);
  glLightfv(GL_LIGHT0,GL_SPECULAR, @m_L0.m_fSpecular);

  glLightfv(GL_LIGHT1,GL_AMBIENT,  @m_L1.m_fAmbient);
  glLightfv(GL_LIGHT1,GL_DIFFUSE,  @m_L1.m_fDiffuse);
  glLightfv(GL_LIGHT1,GL_SPECULAR, @m_L1.m_fSpecular);

  // light attenuation
  glLightf(GL_LIGHT0,GL_QUADRATIC_ATTENUATION, m_L0.m_fQuad_Attenuation);
  glLightf(GL_LIGHT0,GL_LINEAR_ATTENUATION,    m_L0.m_fLin_Attenuation);
  glLightf(GL_LIGHT0,GL_CONSTANT_ATTENUATION,  m_L0.m_fConst_Attenuation);

  glLightf(GL_LIGHT1,GL_QUADRATIC_ATTENUATION, m_L1.m_fQuad_Attenuation);
  glLightf(GL_LIGHT1,GL_LINEAR_ATTENUATION,    m_L1.m_fLin_Attenuation);
  glLightf(GL_LIGHT1,GL_CONSTANT_ATTENUATION,  m_L1.m_fConst_Attenuation);


  // definir material
  ka[1] := 0.0;  ka[2] := 0.0;  ka[3] := 0.0;  ka[4] := 0.0;
  kd[1] := 0.8;  kd[2] := 0.8;  kd[3] := 0.8;  kd[4] := 1.0;
  ks[1] := 0.2;  ks[2] := 0.2;  ks[3] := 0.2;  ks[4] := 1.0;
  ke[1] := 0.0;  ke[2] := 0.0;  ke[3] := 0.0;  ke[4] := 1.0;

  glMaterialfv(GL_FRONT, GL_AMBIENT, @ka);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, @kd);
  glMaterialfv(GL_FRONT, GL_SPECULAR, @ks);
  glMaterialfv(GL_FRONT, GL_EMISSION, @ke);
  glMaterialf(GL_FRONT, GL_SHININESS, 20);

  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  glEnable(GL_NORMALIZE);

end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  SaveConfig;
  LEDCtrl($80);
  SetWiimoteMode(WII_MODE_BASIC);
  bmpWiimote.Free;
end;

procedure TMainForm.HIDCtlDeviceChange(Sender: TObject);
begin
  // see if current device has been removed
  if (CurrentDevice <> nil) and not CurrentDevice.IsPluggedIn then
  begin
    Options1.Enabled := False;
    BatteryTimer.Enabled := False;
    MotionTimer.Enabled := False;
    DrawTimer.Enabled := False;
    RepeatTimer.Enabled := False;
    CursorTimer.Enabled := False;
    HidCtl.CheckIn(CurrentDevice);
  end;

  if (CurrentDevice = nil) and HidCtl.CheckOutByID(CurrentDevice, WII_REMOTE_VID, WII_REMOTE_PID) then
  begin
    SetWiimoteMode(WII_MODE_ACC_IR_EXT);
    Sleep(200);

    if Setting.isNunchukEnabled then
      NunchukStart
    else
      IRSensorStart;

    Sleep(200);

    BatteryRequest;

    Options1.Enabled := True;
    BatteryTimer.Enabled := True;
    if Setting.EnableCursorAtStartup
      then MotionTimer.Enabled := True;
    DrawTimer.Enabled := True;
    RepeatTimer.Enabled := True;
    CursorTimer.Enabled := True;
  end;

  if CurrentDevice = nil then
    MainForm.Caption := 'Wii Remote no detectado'
  else
    MainForm.Caption := 'Form1';
end;

procedure TMainForm.WiiData(HidDev: TJvHidDevice; ReportID: Byte; const Data: Pointer; Size: Word);
var
  Str: string;
begin
  AnalyzePacket(ReportID, PChar(Data), Size);
  Str := Format('R %.2x  ', [ReportID]);
end;

procedure TMainForm.PushKeys(Keys: string);
var
  sc: TShortCut;
  key: Integer;
  winkeyFlag: Boolean;
begin
  winkeyFlag := False;
  if Pos('Win', Keys) > 0 then
  begin
    winkeyFlag := True;
    Keys := Replace(Keys, '+Win+', '');
    Keys := Replace(Keys, 'Win+', '');
    Keys := Replace(Keys, '+Win', '');
    Keys := Replace(Keys, 'Win', '');
  end;
  sc := TextToShortCut(Keys);
  key := sc;

  if sc and scAlt <> 0 then
    Keybd_event(VK_MENU, 0, 0, 0);
  if sc and scCtrl <> 0 then
    Keybd_event(VK_CONTROL, 0, 0, 0);
  if sc and scShift <> 0 then
    Keybd_event(VK_SHIFT, 0, 0, 0);

  if winkeyFlag then
    Keybd_event(VK_LWIN, 0, 0, 0);

  if sc and scAlt <> 0 then
    key := key xor scAlt;
  if sc and scCtrl <> 0 then
    key := key xor scCtrl;
  if sc and scShift <> 0 then
    key := key xor scShift;

  if key <> 0 then
  begin
    Keybd_event(key, 0, 0, 0);
    Keybd_event(key, 0, KEYEVENTF_KEYUP, 0);
  end;

  if winkeyFlag then
    Keybd_event(VK_LWIN, 0, KEYEVENTF_KEYUP, 0);
  if sc and scShift <> 0 then
    Keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
  if sc and scCtrl <> 0 then
    Keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
  if sc and scAlt <> 0 then
    Keybd_event(VK_MENU, 0, KEYEVENTF_KEYUP, 0);

end;
procedure TMainForm.PushXYZ(Data: PChar);
var
  I: Integer;
begin
  for i := High(GraphData) downto 1 do
    GraphData[I] := GraphData[I - 1];

  Wiimote.Sensor.Motion.X:=Byte(Data[2]);
  Wiimote.Sensor.Motion.Y:=Byte(Data[3]);
  Wiimote.Sensor.Motion.Z:=Byte(Data[4]);

  GraphData[0].X := (Wiimote.Sensor.Motion.X - (Setting.Calibration.MotionSensor.X - 128)) div GRAPH_DIV;
  GraphData[0].Y := (Wiimote.Sensor.Motion.Y - (Setting.Calibration.MotionSensor.Y - 128)) div GRAPH_DIV;
  GraphData[0].Z := Wiimote.Sensor.Motion.Z div GRAPH_DIV;
end;
procedure TMainForm.AnalyzePacket(ReportID: Integer; ReportData: PChar; Size: Word);
begin
  case ReportID of
    WII_MODE_ACC:
      PushXYZ(ReportData);
    WII_MODE_ACC_IR:
      begin
        PushXYZ(ReportData);
        AnalyzeIR(ReportData);
      end;
    WII_MODE_ACC_IR_EXT:
      begin
        PushXYZ(ReportData);
        AnalyzeIR(ReportData);
        if Wiimote.isNunchukEnabled then AnalyzeNunchuk(ReportData);
      end;
    WII_MODE_EXPANSION:
    begin
      if(Byte(ReportData[2]) and 2 <> 0)then
      begin
        if Setting.isNunchukEnabled then
          NunchukStart
        else
          IRSensorStart;
        Wiimote.isNunchukEnabled:=true;
      end
      else
      begin
        IRSensorStart;
        Wiimote.isNunchukEnabled:=false;
      end;
      Wiimote.Battery := Integer(ReportData[5]);
    end;
  end;

  AnalyzeButtons((Cardinal(ReportData[0]) shl 8) or Cardinal(ReportData[1]));
end;

procedure TMainForm.AnalyzeIR(ReportData: PChar);
var
  IR_1: Byte;
  IR_2: Byte;
  IR_3: Byte;
  IR_4: Byte;
  IR_5: Byte;
  IR_6: Byte;
begin
  IR_1 := Byte(ReportData[5]); //R
  IR_2 := Byte(ReportData[6]);
  IR_3 := Byte(ReportData[7]);

  IR_4 := Byte(ReportData[8]); //L
  IR_5 := Byte(ReportData[9]);
  IR_6 := Byte(ReportData[10]);

  with Wiimote.Sensor.IR do
  begin
    dot1S := (IR_3 and $0F);
    dot2S := (IR_6 and $0F);

    dot1X := IR_1 + ((IR_3 and $30) shl 4);
    dot1Y := IR_2 + ((IR_3 and $C0) shl 2);

    dot2X := IR_4 + ((IR_6 and $30) shl 4);
    dot2Y := IR_5 + ((IR_6 and $C0) shl 2);

    if (dot2S < $0F) and (dot1S < $0F) then
    begin
      DX := abs((dot2X - dot1X) div 2);
      DY := abs((dot2Y - dot1Y) div 2);

      ScreenX := (dot2X + dot1X) div 2;
      ScreenY := (dot2Y + dot1Y) div 2;

      if dot2X > dot1X then
        LeftIs1 := True
      else
        LeftIs1 := False;
      if dot2Y > dot1Y then
        UpperIs1 := True
      else
        UpperIs1 := False;
    end
    else
    if Wiimote.Sensor.IR.dot2S < $0F then
    begin
      if LeftIs1 then
        ScreenX := dot2X - DX
      else
        ScreenX := dot2X + DX;
      if UpperIs1 then
        ScreenY := dot2Y - DY
      else
        ScreenY := dot2Y + DY;
    end
    else
    if dot1S < $0F then
    begin
      if LeftIs1 then
        ScreenX := dot1X + DX
      else
        ScreenX := dot1X - DX;
      if UpperIs1 then
        ScreenY := dot1Y + DY
      else
        ScreenY := dot1Y - DY;
    end;
  end;
end;

procedure TMainForm.AnalyzeNunchuk(ReportData: PChar);
var
  NU_1: Byte;
  NU_2: Byte;
  NU_3: Byte;
  NU_4: Byte;
  NU_5: Byte;
  NU_6: Byte;
begin
  NU_1 := Byte(ReportData[15]);
  NU_2 := Byte(ReportData[16]);
  NU_3 := Byte(ReportData[17]);
  NU_4 := Byte(ReportData[18]);
  NU_5 := Byte(ReportData[19]);
  NU_6 := Byte(ReportData[20]);

  with Wiimote.Sensor.Nunchuk do
  begin
    Analog.X := NU_1 + (NU_1 xor $17)+$17;
    Analog.Y := NU_2 + (NU_2 xor $17)+$17;

    Motion.X := NU_3 + (NU_3 xor $17)+$17;
    Motion.Y := NU_4 + (NU_4 xor $17)+$17;
    Motion.Z := NU_5 + (NU_5 xor $17)+$17;
  end;

  if NU_6 and $1 <>0 then
    ButtonEvent(WII_BUTTON_Z, WII_EVENT_DOWN, Wiimote.Button.ButtonZ)
  else
    ButtonEvent(WII_BUTTON_Z, WII_EVENT_UP, Wiimote.Button.ButtonZ);

  if NU_6 and $2 <> 0 then
    ButtonEvent(WII_BUTTON_C,WII_EVENT_UP , Wiimote.Button.ButtonC)
  else
    ButtonEvent(WII_BUTTON_C,WII_EVENT_DOWN , Wiimote.Button.ButtonC);

end;

procedure TMainForm.RaiseEvent(ButtonAssign: TButtonAssign; Button: Integer; EventType: Integer;
  var ButtonState: TButtonState; RepeatFlag: Boolean = False);
var
  AssignType: string;
  AssignValue: string;
  StartInfo: TStartupInfo;
  ProcInfo: TProcessInformation;
begin
  AssignType := ButtonAssign.AssignType;
  AssignValue := ButtonAssign.AssignValue;

  if (EventType = WII_EVENT_DOWN) and (RepeatFlag = False) then
    ButtonState.LastPushTime := timegetTime;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_KEY then
  begin
    case EventType of
      WII_EVENT_DOWN: PushKeys(AssignValue);
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT then
  begin
    case EventType of
      WII_EVENT_DOWN: Mouse_Event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0);
      WII_EVENT_UP: Mouse_Event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
          Mouse_Event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_LEFTUP, 0, 0, 0, 0);
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT then
  begin
    case EventType of
      WII_EVENT_DOWN: Mouse_Event(MOUSEEVENTF_RIGHTDOWN, 0, 0, 0, 0);
      WII_EVENT_UP: Mouse_Event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
          Mouse_Event(MOUSEEVENTF_RIGHTDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_RIGHTDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_RIGHTUP, 0, 0, 0, 0);
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE then
  begin
    case EventType of
      WII_EVENT_DOWN: Mouse_Event(MOUSEEVENTF_MIDDLEDOWN, 0, 0, 0, 0);
      WII_EVENT_UP: Mouse_Event(MOUSEEVENTF_MIDDLEUP, 0, 0, 0, 0);
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
          Mouse_Event(MOUSEEVENTF_MIDDLEDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_MIDDLEUP, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_MIDDLEDOWN, 0, 0, 0, 0);
          Mouse_Event(MOUSEEVENTF_MIDDLEUP, 0, 0, 0, 0);
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_WHEEL_UP then
  begin
    case EventType of
      WII_EVENT_DOWN: Mouse_Event(MOUSEEVENTF_WHEEL, 0, 0, WHEEL_FORWARD, 0);
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_WHEEL_DOWN then
  begin
    case EventType of
      WII_EVENT_DOWN: Mouse_Event(MOUSEEVENTF_WHEEL, 0, 0, Cardinal(WHEEL_BACKWARD), 0);
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_WEB_FORWARD then
  begin
    case EventType of
      WII_EVENT_DOWN: PushKeys('Alt+Right');
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_WEB_BACKWARD then
  begin
    case EventType of
      WII_EVENT_DOWN: PushKeys('Alt+Left');
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_VOLUME_UP then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
//          Mixer.GetVolume(0, -1, L, R, M, Stereo, VD, MD, IsSelect);
//          Mixer.SetVolume(0, -1, Min(L + 2048, 65535), Min(R + 2048, 65535), Integer(M <> 0));
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_VOLUME_DOWN then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
//          Mixer.GetVolume(0, -1, L, R, M, Stereo, VD, MD, IsSelect);
//          Mixer.SetVolume(0, -1, Max(L - 2048, 0), Max(R - 2048, 0), Integer(M <> 0));
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_VOLUME_MUTE then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
//          Mixer.GetVolume(0, -1, L, R, M, Stereo, VD, MD, IsSelect);
//          Mixer.SetVolume(0, -1, L, R, Integer(M = 0));
        end;
      WII_EVENT_UP: ;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF then
  begin
    if (EventType = WII_EVENT_DOWN) then
      ButtonPushFlagTrigger := True;
    if (EventType = WII_EVENT_UP) then
      ButtonPushFlagTrigger := False;
    case EventType of
      WII_EVENT_DOWN:
        begin
          CursorOn;
          prevTriggerTime := timeGetTime();
        end;
      WII_EVENT_UP:
        begin
          if (timeGetTime - prevTriggerTime < 300) then
          begin
          //toggled
            if MotionTimerFlag then
              CursorOff
            else
              LEDCtrl($90);
          end;

        end;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_ON then
  begin
    case EventType of
      WII_EVENT_DOWN: CursorOn;
      WII_EVENT_UP: CursorOff;
    end;
  end;

  if AssignType = STR_BUTTON_ASSIGN_TYPE_EXECUTE then
  begin
    case EventType of
      WII_EVENT_DOWN:
        begin
          GetStartupInfo(StartInfo);
          CreateProcess(nil, PChar(AssignValue), nil, nil, False, 0, nil, nil, StartInfo, ProcInfo);
        end;
      WII_EVENT_UP: ;
    end;
  end;

end;

procedure TMainForm.ButtonEvent(Button: Integer; EventType: Integer; var ButtonState: TButtonState);
begin
  if ((EventType = WII_EVENT_UP) and (ButtonState.PushFlag) or (EventType = WII_EVENT_DOWN) and (not ButtonState.PushFlag)) then
  begin
    case Button of
      WII_BUTTON_UP: RaiseEvent(Wiimote.Assign.Up, Button, EventType, ButtonState);
      WII_BUTTON_DOWN: RaiseEvent(Wiimote.Assign.Down, Button, EventType, ButtonState);
      WII_BUTTON_RIGHT: RaiseEvent(Wiimote.Assign.Right, Button, EventType, ButtonState);
      WII_BUTTON_LEFT: RaiseEvent(Wiimote.Assign.Left, Button, EventType, ButtonState);
      // para no dar clique com esquerdo fora da janela do programa
      //WII_BUTTON_A: RaiseEvent(Wiimote.Assign.ButtonA, Button, EventType, ButtonState);
      // para no habilitar o movimento do cursor 2D
      //WII_BUTTON_B: RaiseEvent(Wiimote.Assign.ButtonB, Button, EventType, ButtonState);
      WII_BUTTON_HOME: RaiseEvent(Wiimote.Assign.Home, Button, EventType, ButtonState);
      WII_BUTTON_PLUS: RaiseEvent(Wiimote.Assign.Plus, Button, EventType, ButtonState);
      WII_BUTTON_MINUS: RaiseEvent(Wiimote.Assign.Minus, Button, EventType, ButtonState);
      WII_BUTTON_1: RaiseEvent(Wiimote.Assign.Button1, Button, EventType, ButtonState);
      WII_BUTTON_2: RaiseEvent(Wiimote.Assign.Button2, Button, EventType, ButtonState);
      WII_BUTTON_C: RaiseEvent(Wiimote.Assign.ButtonC, Button, EventType, ButtonState);
      WII_BUTTON_Z: RaiseEvent(Wiimote.Assign.ButtonZ, Button, EventType, ButtonState);
    end;
  end;
  if (EventType = WII_EVENT_DOWN) then
    ButtonState.PushFlag := True;
  if (EventType = WII_EVENT_UP) then
    ButtonState.PushFlag := False;
end;

procedure TMainForm.RefreshButtonStatus();
begin
  StatusBar1.Panels[1].Text:='Bateria: '+Format('%.f', [(Wiimote.Battery / 192) *100])+'%';
end;

procedure TMainForm.AnalyzeButtons(Buttons: Cardinal);
begin
  if Buttons and WII_BUTTON_UP <> 0 then
    ButtonEvent(WII_BUTTON_UP, WII_EVENT_DOWN, Wiimote.Button.Up)
  else
    ButtonEvent(WII_BUTTON_UP, WII_EVENT_UP, Wiimote.Button.Up);
  if Buttons and WII_BUTTON_DOWN <> 0 then
    ButtonEvent(WII_BUTTON_DOWN, WII_EVENT_DOWN, Wiimote.Button.Down)
  else
    ButtonEvent(WII_BUTTON_DOWN, WII_EVENT_UP, Wiimote.Button.Down);
  if Buttons and WII_BUTTON_RIGHT <> 0 then
    ButtonEvent(WII_BUTTON_RIGHT, WII_EVENT_DOWN, Wiimote.Button.Right)
  else
    ButtonEvent(WII_BUTTON_RIGHT, WII_EVENT_UP, Wiimote.Button.Right);
  if Buttons and WII_BUTTON_LEFT <> 0 then
    ButtonEvent(WII_BUTTON_LEFT, WII_EVENT_DOWN, Wiimote.Button.Left)
  else
    ButtonEvent(WII_BUTTON_LEFT, WII_EVENT_UP, Wiimote.Button.Left);
  if Buttons and WII_BUTTON_HOME <> 0 then
    ButtonEvent(WII_BUTTON_HOME, WII_EVENT_DOWN, Wiimote.Button.Home)
  else
    ButtonEvent(WII_BUTTON_HOME, WII_EVENT_UP, Wiimote.Button.Home);
  if Buttons and WII_BUTTON_MINUS <> 0 then
    ButtonEvent(WII_BUTTON_MINUS, WII_EVENT_DOWN, Wiimote.Button.Minus)
  else
    ButtonEvent(WII_BUTTON_MINUS, WII_EVENT_UP, Wiimote.Button.Minus);
  if Buttons and WII_BUTTON_PLUS <> 0 then
    ButtonEvent(WII_BUTTON_PLUS, WII_EVENT_DOWN, Wiimote.Button.Plus)
  else
    ButtonEvent(WII_BUTTON_PLUS, WII_EVENT_UP, Wiimote.Button.Plus);
  if Buttons and WII_BUTTON_A <> 0 then
    ButtonEvent(WII_BUTTON_A, WII_EVENT_DOWN, Wiimote.Button.ButtonA)
  else
    ButtonEvent(WII_BUTTON_A, WII_EVENT_UP, Wiimote.Button.ButtonA);
  if Buttons and WII_BUTTON_B <> 0 then
    ButtonEvent(WII_BUTTON_B, WII_EVENT_DOWN, Wiimote.Button.ButtonB)
  else
    ButtonEvent(WII_BUTTON_B, WII_EVENT_UP, Wiimote.Button.ButtonB);
  if Buttons and WII_BUTTON_1 <> 0 then
    ButtonEvent(WII_BUTTON_1, WII_EVENT_DOWN, Wiimote.Button.Button1)
  else
    ButtonEvent(WII_BUTTON_1, WII_EVENT_UP, Wiimote.Button.Button1);
  if Buttons and WII_BUTTON_2 <> 0 then
    ButtonEvent(WII_BUTTON_2, WII_EVENT_DOWN, Wiimote.Button.Button2)
  else
    ButtonEvent(WII_BUTTON_2, WII_EVENT_UP, Wiimote.Button.Button2);
end;

procedure TMainForm.SetWiimoteMode(Mode:Byte);
var
  Buf: array[0..64] of Byte;
begin
  Buf[0] := $12;
  Buf[1] := $00;
  Buf[2] := Mode;

  SendToDevice(Buf, 3);
end;

procedure TMainForm.LEDCtrl(LEDState: Integer);
var
  Buf: array[0..64] of Byte;
begin
  Buf[0] := $11;
  Buf[1] := LEDState;
  Buf[2] := $00;
  SendToDevice(Buf, 2);
end;

procedure TMainForm.BatteryRequest();
var
  Buf: array[0..64] of Byte;
begin
  Buf[0] := $15;
  Buf[1] := $00;
  Buf[2] := $00;
  SendToDevice(Buf, 2);
end;

procedure TMainForm.IRSensorStart();
var
  Buf: array[0..64] of Byte;
begin
  SetWiimoteMode(WII_MODE_ACC_IR);

  Buf[0] := $13;
  Buf[1] := $04;
  SendToDevice(Buf, 1);

  Buf[0] := $1A;
  Buf[1] := $04;
  SendToDevice(Buf, 1);

  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $B0;
  Buf[3] := $00;
  Buf[4] := $30;
  Buf[5] := $01;
  Buf[6] := $08;
  SendToDevice(Buf, 6);

  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $B0;
  Buf[3] := $00;
  Buf[4] := $06;
  Buf[5] := $01;
  Buf[6] := $90;
  SendToDevice(Buf, 6);

  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $B0;
  Buf[3] := $00;
  Buf[4] := $08;
  Buf[5] := $01;
  Buf[6] := $C0;
  SendToDevice(Buf, 6);

  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $B0;
  Buf[3] := $00;
  Buf[4] := $1A;
  Buf[5] := $01;
  Buf[6] := $40;
  SendToDevice(Buf, 6);

  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $B0;
  Buf[3] := $00;
  Buf[4] := $33;
  Buf[5] := $01;
  Buf[6] := $33;
  SendToDevice(Buf, 6);
end;

procedure TMainForm.NunchukStart();
var
  Buf: array[0..64] of Byte;
begin
  SetWiimoteMode(WII_MODE_ACC_IR_EXT);
  Buf[0] := $16;
  Buf[1] := $04;
  Buf[2] := $A4;
  Buf[3] := $00;
  Buf[4] := $40;
  Buf[5] := $01;
  SendToDevice(Buf, 5);
end;
procedure TMainForm.SendToDevice(Buf: array of Byte; WriteLength: Integer);
var
  numToWrite: Cardinal;
  numWritten: Cardinal;
  numErr: DWORD;
begin
  if not Assigned(CurrentDevice) then
  begin
    MainForm.Caption := 'Wii Remote no encontrado';
    Exit;
  end;

  numToWrite := CurrentDevice.Caps.OutputReportByteLength;

  CurrentDevice.OpenFile;
  if not CurrentDevice.WriteFile(Buf, numToWrite, numWritten) then
  begin
    numErr := GetLastError;
    MainForm.Caption := 'Erro(' + IntToStr(numErr) + '):' + SysErrorMessage(numErr);
  end
  else
  begin
    CurrentDevice.CloseFile;
  end;
end;

procedure TMainForm.DrawTimerTimer(Sender: TObject);
begin
  RefreshButtonStatus;
  if Setting.DrawGraph then DrawImage();
end;

procedure TMainForm.MotionTimerTimer(Sender: TObject);

  function GetSpeedFast(SpeedFast: Integer): Extended;
  begin
    Result := 2.25 - (SpeedFast * 0.1);
  end;

  function GetSpeedSlow(SpeedSlow: Integer): Extended;
  begin
    Result := 12 - SpeedSlow;
  end;

  function GetY(Y: Integer): Extended;
  begin
    Result:=768/(768-Setting.DisplayEdgeAdjustment*2)*Y-Setting.DisplayEdgeAdjustment;
  end;
var
  Point: TPoint;
  dx, dy: Integer;
  x, y: Integer;
  axn, ayn: Integer;
begin
  GetCursorPos(Point);

  if ((Setting.CursorMode = STR_CURSOR_MODE_MOTION) or ((Setting.CursorMode = STR_CURSOR_MODE_IR) and ((Wiimote.Sensor.IR.dot2S = $0F) and (Wiimote.Sensor.IR.dot1S = $0F)) and Setting.UseMotionWhenLostIR)) then
  begin
    if abs(Wiimote.Sensor.Motion.X - Setting.Calibration.MotionSensor.X) < 2 then
      dx := 0
    else
      dx := Wiimote.Sensor.Motion.X - Setting.Calibration.MotionSensor.X;
    if abs(Wiimote.Sensor.Motion.Y - Setting.Calibration.MotionSensor.Y) < 2 then
      dy := 0
    else
      dy := Wiimote.Sensor.Motion.Y - Setting.Calibration.MotionSensor.Y;

    if Setting.CursorSpeed = STR_CURSOR_SPEED_FASTER then
    begin
      if not ButtonPushFlagTrigger then
      begin
        dx := Ceil(dx / GetSpeedSlow(Setting.MotionSpeedSlow) * 1.2);
        dy := Ceil(dy / GetSpeedSlow(Setting.MotionSpeedSlow));
      end
      else
      begin
        dx := trunc(dx / GetSpeedFast(Setting.MotionSpeedFast) * 1.2);
        dy := trunc(dy / GetSpeedFast(Setting.MotionSpeedFast));
      end;
    end
    else
    if Setting.CursorSpeed = STR_CURSOR_SPEED_SLOWER then
    begin
      if ButtonPushFlagTrigger then
      begin
        dx := Ceil(dx / GetSpeedSlow(Setting.MotionSpeedSlow) * 1.2);
        dy := Ceil(dy / GetSpeedSlow(Setting.MotionSpeedSlow));
      end
      else
      begin
        if abs(Wiimote.Sensor.Motion.X - Setting.Calibration.MotionSensor.X) < 2 then
          dx := 0
        else
          dx := Wiimote.Sensor.Motion.X - Setting.Calibration.MotionSensor.X;
        if abs(Wiimote.Sensor.Motion.Y - Setting.Calibration.MotionSensor.Y) < 2 then
          dy := 0
        else
          dy := Wiimote.Sensor.Motion.Y - Setting.Calibration.MotionSensor.Y;
        dx := trunc(dx / GetSpeedFast(Setting.MotionSpeedFast) * 1.2);
        dy := trunc(dy / GetSpeedFast(Setting.MotionSpeedFast));
      end;
    end;
    if Setting.InvertH then
      dx := -dx;
    if Setting.InvertV then
      dy := -dy;
    SetCursorPos(Point.X + dx, Point.Y + dy);
  end
  else if Setting.CursorMode = STR_CURSOR_MODE_IR then
  begin

    x := roundoff(Screen.Width * (1024 - Wiimote.Sensor.IR.ScreenX) / 1024);
    y := roundoff(Screen.Height * (GetY(Wiimote.Sensor.IR.ScreenY) / 768));

    x := (x + Point.X * 5) div 6;
    y := (y + Point.Y * 5) div 6;

    SetCursorPos(x, y);
  end;

  if Setting.CursorMode= STR_CURSOR_MODE_NUNCHUK_MOTION then
  begin
      {
        axn:=0;
        ayn:=0;
        if (Wiimote.Sensor.Nunchuk.X > $90) then axn := (Wiimote.Sensor.Nunchuk.X - $90)div 8;
        if (Wiimote.Sensor.Nunchuk.X < $70) then axn := (Wiimote.Sensor.Nunchuk.X - $70)div 8;
        if (Wiimote.Sensor.Nunchuk.Y > $90) then ayn := (Wiimote.Sensor.Nunchuk.Y - $90)div 8;
        if (Wiimote.Sensor.Nunchuk.Y < $70) then ayn := (Wiimote.Sensor.Nunchuk.Y - $70)div 8;
        GetCursorPos(Point);
        x := Point.x;
        y := Point.y;
        x :=x+axn;
        y :=y-ayn;
        SetCursorPos(x,y);
        }
  end;

  if Setting.CursorMode= STR_CURSOR_MODE_NUNCHUK_ANALOG then
  begin
    if Wiimote.isNunchukEnabled then
    begin
      axn:=0;
      ayn:=0;
      if (Wiimote.Sensor.Nunchuk.Analog.X > $110) then axn := (Wiimote.Sensor.Nunchuk.Analog.X - $110)div (22-Setting.AnalogCursorSpeed);
      if (Wiimote.Sensor.Nunchuk.Analog.X < $90) then axn := (Wiimote.Sensor.Nunchuk.Analog.X - $90)div (22-Setting.AnalogCursorSpeed);
      if (Wiimote.Sensor.Nunchuk.Analog.Y > $110) then ayn := (Wiimote.Sensor.Nunchuk.Analog.Y - $110)div (22-Setting.AnalogCursorSpeed);
      if (Wiimote.Sensor.Nunchuk.Analog.Y < $90) then ayn := (Wiimote.Sensor.Nunchuk.Analog.Y - $90)div (22-Setting.AnalogCursorSpeed);
      GetCursorPos(Point);
      x := Point.x;
      y := Point.y;

      if Setting.NunchukAnalogInvertH then axn := -axn;
      if Setting.NunchukAnalogInvertV then ayn := -ayn;

      x :=x+axn;
      y :=y-ayn;

      SetCursorPos(x,y);
    end;
  end;

  GetCursorPos(Point);
  if IsDisplayEdge(Point) and Setting.RumbleOnEdge then
  begin
    //stop after 60ms
    if RumbleFlag and ((timeGetTime - LastRumble) > 60) then
    begin
      LEDCtrl($90);
      RumbleFlag := False;
    end
    else
    if (RumbleFlag = False) and not PrevEdged then
    begin
      LEDCtrl($91);
      LastRumble := timeGetTime;
      RumbleFlag := True;
    end;
    PrevEdged := True;
  end
  else
  begin
    PrevEdged := False;
    if RumbleFlag and ((timeGetTime - LastRumble) > 60) then
    begin
      LEDCtrl($90);
      RumbleFlag := False;
    end
  end;
end;

procedure TMainForm.RepeatTimerTimer(Sender: TObject);
var
  t: TButtonState;
begin
  if Wiimote.Button.Up.PushFlag and IsEventRepeat(Wiimote.Assign.Up.AssignType) and (timeGetTime - Wiimote.Button.Up.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Up, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Down.PushFlag and IsEventRepeat(Wiimote.Assign.Down.AssignType) and (timeGetTime - Wiimote.Button.Down.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Down, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Right.PushFlag and IsEventRepeat(Wiimote.Assign.Right.AssignType) and (timeGetTime - Wiimote.Button.Right.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Right, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Left.PushFlag and IsEventRepeat(Wiimote.Assign.Left.AssignType) and (timeGetTime - Wiimote.Button.Left.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Left, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.ButtonA.PushFlag and IsEventRepeat(Wiimote.Assign.ButtonA.AssignType) and (timeGetTime - Wiimote.Button.ButtonA.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.ButtonA, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.ButtonB.PushFlag and IsEventRepeat(Wiimote.Assign.ButtonB.AssignType) and (timeGetTime - Wiimote.Button.ButtonB.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.ButtonB, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Plus.PushFlag and IsEventRepeat(Wiimote.Assign.Plus.AssignType) and (timeGetTime - Wiimote.Button.Plus.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Plus, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Minus.PushFlag and IsEventRepeat(Wiimote.Assign.Minus.AssignType) and (timeGetTime - Wiimote.Button.Minus.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Minus, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Home.PushFlag and IsEventRepeat(Wiimote.Assign.Home.AssignType) and (timeGetTime - Wiimote.Button.Home.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Home, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Button1.PushFlag and IsEventRepeat(Wiimote.Assign.Button1.AssignType) and (timeGetTime - Wiimote.Button.Button1.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Button1, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.Button2.PushFlag and IsEventRepeat(Wiimote.Assign.Button2.AssignType) and (timeGetTime - Wiimote.Button.Button2.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.Button2, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.ButtonC.PushFlag and IsEventRepeat(Wiimote.Assign.ButtonC.AssignType) and (timeGetTime - Wiimote.Button.ButtonC.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.ButtonC, 0, WII_EVENT_DOWN, t, True);
  if Wiimote.Button.ButtonZ.PushFlag and IsEventRepeat(Wiimote.Assign.ButtonZ.AssignType) and (timeGetTime - Wiimote.Button.ButtonZ.LastPushTime > 800) then
    RaiseEvent(Wiimote.Assign.ButtonZ, 0, WII_EVENT_DOWN, t, True);
end;


procedure TMainForm.CalibrateButtonClick(Sender: TObject);
begin
  // define o "zero" relativo da acelerao
  Setting.Calibration.MotionSensor := Wiimote.Sensor.Motion;
end;

procedure TMainForm.ShowMenuClick(Sender: TObject);
begin
  Application.Restore;
end;

procedure TMainForm.TrayIcon1DblClick(Sender: TObject);
begin
  Application.Restore;
end;

procedure TMainForm.ExitMenuClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.BatteryTimerTimer(Sender: TObject);
begin
  BatteryRequest;
end;

procedure TMainForm.IRResetButtonClick(Sender: TObject);
begin
  IRSensorStart;
end;

procedure TMainForm.DrawImage();
var
  i: Integer;
begin
  // usa botes direcionais do WiiRemote para manipular
  // o MTK_Orbit e girar a cmera
  if Wiimote.Button.Up.PushFlag then
    begin
    MTK_Orbit(0,-1);
    end;
  if Wiimote.Button.Down.PushFlag then
    begin
    MTK_Orbit(0,+1);
    end;
  if Wiimote.Button.Right.PushFlag then
    begin
    MTK_Orbit(+1,0);
    end;
  if Wiimote.Button.Left.PushFlag then
    begin
    MTK_Orbit(-1,0);
    end;


  if Wiimote.Button.ButtonA.PushFlag then
    begin
    CalculaBase;
    end;

  if Wiimote.Button.ButtonB.PushFlag then
    begin

    end;

  if Wiimote.Button.Home.PushFlag then
    begin
    // zera a posio do cursor
    for i := 1 to 3 do
      pontos[i] := 0;
    end;

  if Wiimote.Button.Plus.PushFlag then
    begin
    // ajusta o fator do movimento do cursor
    inc(fator,10);
    end;

  if Wiimote.Button.Minus.PushFlag then
    begin
    // ajusta o fator do movimento do cursor
    dec(fator,10);
    if fator <= 0 then fator := 10;
    end;

  if Wiimote.Button.Button1.PushFlag then
    begin
    MTK_Dolly(+1);
    end;
  if Wiimote.Button.Button2.PushFlag then
    begin
    MTK_Dolly(-1);
    end;

end;

procedure TMainForm.CursorOn();
begin
  MotionTimerFlag := MotionTimer.Enabled;
  MotionTimer.Enabled := True;
  LEDCtrl($90);
  if Setting.ReCalibrate then
    CalibrateButtonClick(nil);
end;

procedure TMainForm.CursorOff();
begin
  MotionTimerFlag := False;
  MotionTimer.Enabled := False;
  LEDCtrl($80);
  if Setting.ReCalibrate then
    CalibrateButtonClick(nil);
end;

procedure TMainForm.Exit1Click(Sender: TObject);
begin
  Close;
end;


procedure TMainForm.CursorTimerTimer(Sender: TObject);

  procedure SetAccel(AssignType: string; var ax: Integer; var ay: Integer);
  begin
    if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_UP then
      ay := 1;
    if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_DOWN then
      ay := -1;
    if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT then
      ax := 1;
    if AssignType = STR_BUTTON_ASSIGN_TYPE_CURSOR_LEFT then
      ax := -1;
  end;

var
  i, ax, ay: Integer;
  mf: TMatriz4x4;
  da: TVet3d;
  massa: TdMass;
begin
  ax := 0;
  ay := 0;
  if Wiimote.Button.Up.PushFlag and IsCursorEvent(Wiimote.Assign.Up.AssignType) then
    SetAccel(Wiimote.Assign.Up.AssignType, ax, ay);
  if Wiimote.Button.Down.PushFlag and IsCursorEvent(Wiimote.Assign.Down.AssignType) then
    SetAccel(Wiimote.Assign.Down.AssignType, ax, ay);
  if Wiimote.Button.Right.PushFlag and IsCursorEvent(Wiimote.Assign.Right.AssignType) then
    SetAccel(Wiimote.Assign.Right.AssignType, ax, ay);
  if Wiimote.Button.Left.PushFlag and IsCursorEvent(Wiimote.Assign.Left.AssignType) then
    SetAccel(Wiimote.Assign.Left.AssignType, ax, ay);
  if Wiimote.Button.ButtonA.PushFlag and IsCursorEvent(Wiimote.Assign.ButtonA.AssignType) then
    SetAccel(Wiimote.Assign.ButtonA.AssignType, ax, ay);
  if Wiimote.Button.ButtonB.PushFlag and IsCursorEvent(Wiimote.Assign.ButtonB.AssignType) then
    SetAccel(Wiimote.Assign.ButtonB.AssignType, ax, ay);
  if Wiimote.Button.Plus.PushFlag and IsCursorEvent(Wiimote.Assign.Plus.AssignType) then
    SetAccel(Wiimote.Assign.Plus.AssignType, ax, ay);
  if Wiimote.Button.Minus.PushFlag and IsCursorEvent(Wiimote.Assign.Minus.AssignType) then
    SetAccel(Wiimote.Assign.Minus.AssignType, ax, ay);
  if Wiimote.Button.Home.PushFlag and IsCursorEvent(Wiimote.Assign.Home.AssignType) then
    SetAccel(Wiimote.Assign.Home.AssignType, ax, ay);
  if Wiimote.Button.Button1.PushFlag and IsCursorEvent(Wiimote.Assign.Button1.AssignType) then
    SetAccel(Wiimote.Assign.Button1.AssignType, ax, ay);
  if Wiimote.Button.Button2.PushFlag and IsCursorEvent(Wiimote.Assign.Button2.AssignType) then
    SetAccel(Wiimote.Assign.Button2.AssignType, ax, ay);
  if Wiimote.Button.ButtonC.PushFlag and IsCursorEvent(Wiimote.Assign.ButtonC.AssignType) then
    SetAccel(Wiimote.Assign.ButtonC.AssignType, ax, ay);
  if Wiimote.Button.ButtonZ.PushFlag and IsCursorEvent(Wiimote.Assign.ButtonZ.AssignType) then
    SetAccel(Wiimote.Assign.ButtonZ.AssignType, ax, ay);

  for i := 0 to Length(objetosCena.Objetos)-1 do
    if objetosCena.Objetos[i].selecionado then
      dBodyAddForce(Body[i],0,0,0);
      
  // se tinha pina e agora no tem mais
  if pinca and not(Wiimote.Button.ButtonA.PushFlag and Wiimote.Button.ButtonB.PushFlag) then
    begin
    // soltou a pina
    for i := 0 to Length(objetosCena.Objetos)-1 do
      if objetosCena.Objetos[i].selecionado then
        begin
        c3d.a[1] := Wiimote.Sensor.Motion.X;
        c3d.a[2] := Wiimote.Sensor.Motion.Y;
        c3d.a[3] := Wiimote.Sensor.Motion.Z;

        // movimento em uvn
        mf := matInv(matRotacao(VPN,VUP));

        // variao da acelerao entre frames
        da[1] := +CorrigeMovimento[1]*(c3d.a[1]-c3d.aAnt[1]);
        da[3] := -CorrigeMovimento[2]*(c3d.a[2]-c3d.aAnt[2]);
        da[2] := +CorrigeMovimento[3]*(c3d.a[3]-c3d.aAnt[3]);
        da := prodMatVet(mf,da);

        dBodySetLinearVel(Body[i], da[1], da[2], da[3]);
        dBodyGetMass(Body[i],massa);

        c3d.aAnt[1] := Wiimote.Sensor.Motion.X;
        c3d.aAnt[2] := Wiimote.Sensor.Motion.Y;
        c3d.aAnt[3] := Wiimote.Sensor.Motion.Z;
        end;
    end;

  // sinaliza se a pina est ativa quando os botes A e B esto pressionados
  if pinca then
    begin
      soltou := false;
      // pina estava ativa, mas agora soltou os dedos
      if not (Wiimote.Button.ButtonA.PushFlag and Wiimote.Button.ButtonB.PushFlag)
        then soltou := true;
    end
    else
      soltou := false;

  pinca := Wiimote.Button.ButtonA.PushFlag and Wiimote.Button.ButtonB.PushFlag;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
FinalizaOpenGL;
ShutDownODE;
end;

function RndAng(x: real): real;
begin
if frac(x/10) < 0.5
  then Result := Round(x/10)*10
  else Result := (Round(x/10) + 0.5)*10;
end;

procedure TMainForm.TimerGLTimer(Sender: TObject);
var fi, teta, teta1, z, rotcam, rotaccel, DX, DY, tamQuadGradeX, tamQuadGradeY: real;
i,j,c, QuantQuadGrade: integer;
ti,tf:TTime;
m1,m2,m3,m4,m5,mf,mc: TMatriz4x4;
vaux, da, movimento: Tvet3d;
matproj: array[0..15] of Single;
begin
  // Atualiza ODE
  UpdateODE;
  // atualiza objetos wii
  for i := 0 to Length(objetosCena.Objetos)-1 do
    begin
    if Body[i] <> nil then
      begin
      if objetosCena.Objetos[i].ODEMove then
        begin
        objetosCena.Objetos[i].pant[1] := objetosCena.Objetos[i].p[1];
        objetosCena.Objetos[i].pant[2] := objetosCena.Objetos[i].p[2];
        objetosCena.Objetos[i].pant[3] := objetosCena.Objetos[i].p[3];

        objetosCena.Objetos[i].p[1] := dBodyGetPosition(Body[i])[0];
        objetosCena.Objetos[i].p[2] := dBodyGetPosition(Body[i])[1];
        objetosCena.Objetos[i].p[3] := dBodyGetPosition(Body[i])[2];

        objetosCena.Objetos[i].v[1] := dBodyGetLinearVel(Body[i])[0];
        objetosCena.Objetos[i].v[2] := dBodyGetLinearVel(Body[i])[1];
        objetosCena.Objetos[i].v[3] := dBodyGetLinearVel(Body[i])[2];

        objetosCena.Objetos[i].a[1] := dBodyGetForce(Body[i])[0];
        objetosCena.Objetos[i].a[2] := dBodyGetForce(Body[i])[1];
        objetosCena.Objetos[i].a[3] := dBodyGetForce(Body[i])[2];

        objetosCena.Objetos[i].Q[0] := dBodyGetQuaternion(Body[i])[0];
        objetosCena.Objetos[i].Q[1] := dBodyGetQuaternion(Body[i])[1];
        objetosCena.Objetos[i].Q[2] := dBodyGetQuaternion(Body[i])[2];
        objetosCena.Objetos[i].Q[3] := dBodyGetQuaternion(Body[i])[3];

        // grava o rastro de cada obj
        if objetosCena.Objetos[i].MostraRastroSempre {or objetosCena.Objetos[i].MostraRastroSolto} then
          begin
          SetLength(objetosCena.Objetos[i].rastro, length(objetosCena.Objetos[i].rastro)+1);
          objetosCena.Objetos[i].rastro[length(objetosCena.Objetos[i].rastro)-1] := objetosCena.Objetos[i].p;
          end;
        end
        else
        begin
        dBodySetPosition(Body[i],objetosCena.Objetos[i].p[1],objetosCena.Objetos[i].p[2],objetosCena.Objetos[i].p[3]);
        dBodySetQuaternion(Body[i],objetosCena.Objetos[i].Q);
        end;
      end;
    end;

  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(60,Panel1.Width/Panel1.Height,1,1000);
  
  for i := 1 to 3 do
    VPN[i] := {pontos[i]} - VRP[i];
  gluLookAt(VRP[1],VRP[2],VRP[3],ALVO[1],ALVO[2],ALVO[3],VUP[1],VUP[2],VUP[3]);

  glGetFloatv(GL_PROJECTION_MATRIX, @matproj);
  mc[1,1] := matProj[0];
  mc[2,1] := matProj[1];
  mc[3,1] := matProj[2];
  mc[4,1] := matProj[3];
  mc[1,2] := matProj[4];
  mc[2,2] := matProj[5];
  mc[3,2] := matProj[6];
  mc[4,2] := matProj[7];
  mc[1,3] := matProj[8];
  mc[2,3] := matProj[9];
  mc[3,3] := matProj[10];
  mc[4,3] := matProj[11];
  mc[1,4] := matProj[12];
  mc[2,4] := matProj[13];
  mc[3,4] := matProj[14];
  mc[4,4] := matProj[15];

  glMatrixMode(GL_MODELVIEW);

  c3d.a[1] := Wiimote.Sensor.Motion.X;
  c3d.a[2] := Wiimote.Sensor.Motion.Y;
  c3d.a[3] := Wiimote.Sensor.Motion.Z;

  t := MilliSecondsBetween(now,tempoA)/1000;
  if (Wiimote.Button.ButtonA.PushFlag and (Wiimote.Sensor.IR.dot1S>0)) or (Wiimote.Button.ButtonA.PushFlag and (Wiimote.Sensor.IR.dot2S>0))then
    begin

      // movimento em uvn
      mf := matInv(matRotacao(VPN,VUP));

      // variao da acelerao entre frames
      da[1] := +CorrigeMovimento[1]*(c3d.a[1]-c3d.aAnt[1]);
      da[3] := -CorrigeMovimento[2]*(c3d.a[2]-c3d.aAnt[2]);
      da[2] := +CorrigeMovimento[3]*(c3d.a[3]-c3d.aAnt[3]);
       
      // calcula variao da acelerao corrigida pela cmera

      da := prodMatVet(mf,da);

      // grava rastro do cursor
      SetLength(rastro,length(rastro)+1);
      for i := 1 to 3 do
        rastro[length(rastro)-1,i] := pontos[i];
    end
    else
    begin
      for i := 1 to 3 do
        begin
        vpontos[i]:=0;
        da[i] := 0;
        end;
      SetLength(rastro,0);
      CalibrateButtonClick(Self);
    end;

  // calcula nova posio a partir da variao da acelerao entre os frames
  // com correo da cmera, se uvn ativo
  for i := 1 to 3 do
    begin
    // correo movimento em Y
    //da[i] := da[i]*CorrigeMovimento[i];

    vpontos[i]:=vpontos[i]+fator*(da[i])*t;
    pontos[i]:=pontos[i]+vpontos[i]*t+fator*(da[i])*sqr(t)/2;
    end;

  c3d.p := pontos;

  CoordenadasGrade(pontos);

  // desenha rastro do cursor
  glColor3f(1,1,0);
  glBegin(GL_LINE_STRIP);
    for i := 0 to Length(rastro)-1 do
      glVertex3f(-rastro[i,1],-rastro[i,2],-rastro[i,3]);
  glEnd;

  // desenha rastro dos objetos
  for i := 0 to Length(objetosCena.Objetos)-1 do
    begin
    glLineWidth(3);
    glColor3f(objetosCena.Objetos[i].cor.R/2,objetosCena.Objetos[i].cor.G/2,objetosCena.Objetos[i].cor.B/2);
    glBegin(GL_LINE_STRIP);
      for j := 0 to length(objetosCena.Objetos[i].rastro)-1 do
        glVertex3f(objetosCena.Objetos[i].rastro[j,1],objetosCena.Objetos[i].rastro[j,2],objetosCena.Objetos[i].rastro[j,3]);
    glEnd;
    glLineWidth(1);
    end;


  //fi := 10*RadToDeg(ArcTan(pontos['z']));
  //teta := 10*RadToDeg(ArcTan(pontos['y']));
  //teta1 := 10*RadToDeg(ArcTan(pontos['x']));

  Eixos(5,true);


  {fi := ArcTan2(sqrt(sqr(Wiimote.Sensor.Motion.X-125)+sqr(Wiimote.Sensor.Motion.Y-127)),(Wiimote.Sensor.Motion.Z-125));
  Label3.Caption := FloatToStr(fi);
  Label7.Caption := FloatToStr(teta-teta1);
  Label8.Caption := FloatToStr(Wiimote.Sensor.Motion.X-125);
  Label9.Caption := FloatToStr(Wiimote.Sensor.Motion.Y-127);
  Label10.Caption := FloatToStr(Wiimote.Sensor.Motion.Z-125);}
  {gx = g * sin(fi) * cos(t)
  gy = g * sin(fi) * sin(t)
  gz = g * cos(fi)}


  // *****************************************************************************
  // Motion analysis
  // http://wiibrew.org/wiki/Wiimote/Motion_analysis
  // *****************************************************************************
  //  Calibrar
  //  1) Horizontal with the A button facing up: (x1, y1, z1)
  //  2) IR sensor down on the table so the expansion port is facing up: (x2, y2, z2)
  //  3) Laying on its side, so the left side is facing up: (x3, y3, z3)
  //  x0 = (x1 + x2)/2
  //  y0 = (y1 + y3)/2
  //  z0 = (z2 + z3)/2
  //
  //  x := (Wiimote.Sensor.Motion.x-x0)/(x3-x0);
  //  y := (Wiimote.Sensor.Motion.y-y0)/(y2-y0);
  //  z := (Wiimote.Sensor.Motion.z-z0)/(z1-z0);

  // *****************************************************************************
  // Estimativa de distncia at a WiiBar pela Cmera
  // http://wiiphysics.site88.net/physics.html
  // *****************************************************************************
  // z := dLEDbarra / (2*Tan(((DegToRad(HFOV)/1024 + DegToRad(VFOV)/768)*sqrt(sqr(Wiimote.Sensor.IR.dot1X-Wiimote.Sensor.IR.dot2X)+sqr(Wiimote.Sensor.IR.dot1Y-Wiimote.Sensor.IR.dot2Y)))/4));
  // LbDistLED.Caption := 'Wii Bar: ' + FloatToStrF(z,ffFixed,6,2) + ' m';


  AngEuler;

  // *****************************************************************************
  // Rotao pela acelerao
  // http://wiibrew.org/wiki/Wiimote/Pointing
  // *****************************************************************************
  // colocar a cmera para baixo e calibrar
  rotaccel := ArcTan2(Wiimote.Sensor.Motion.Z-Setting.Calibration.MotionSensor.Z, Wiimote.Sensor.Motion.X-Setting.Calibration.MotionSensor.X) - (PI / 2.0);
  if rotaccel < 0
    then rotaccel := 2*pi-abs(rotaccel);

  // *****************************************************************************
  // Rotao pela cmera
  // *****************************************************************************
  DX := abs(Wiimote.Sensor.IR.dot1Y-Wiimote.Sensor.IR.dot2Y);
  DY := abs(Wiimote.Sensor.IR.dot2X-Wiimote.Sensor.IR.dot1X);
  if Wiimote.Sensor.IR.LeftIs1 and Wiimote.Sensor.IR.UpperIs1
    then rotcam := +ArcTan2(DY, DX)+3*pi/2;
  if Wiimote.Sensor.IR.LeftIs1 and not Wiimote.Sensor.IR.UpperIs1
    then rotcam := -ArcTan2(DY, DX)+1*pi/2;
  if not Wiimote.Sensor.IR.LeftIs1 and not Wiimote.Sensor.IR.UpperIs1
    then rotcam := +ArcTan2(DY, DX)+1*pi/2;
  if not Wiimote.Sensor.IR.LeftIs1 and Wiimote.Sensor.IR.UpperIs1
    then rotcam := -ArcTan2(DY, DX)+3*pi/2;


  // desenha controle na cena
  glColor3f(1,1,0);
  glPushMatrix;
    glTranslate(-pontos[1],-pontos[2],-pontos[3]);
    //glRotate(-fi,1,0,0);
    //glRotate(teta1,0,1,0);
    //glRotate(RadToDeg(gama),0,0,1);
    glScale(1,3,1);
    cubo(GL_QUADS);
  glPopMatrix;


  // Desenha cho (com transparncia)
  glPushMatrix;
    QuantQuadGrade := 20;
    tamQuadGradeX := vFloorBox.Size.x / QuantQuadGrade;
    tamQuadGradeY := vFloorBox.Size.y / QuantQuadGrade;
    DesenhaChao(tamQuadGradeX,tamQuadGradeY,QuantQuadGrade);
  glPopMatrix;

  // desenhar objetos da cena
  for i := 1 to 3 do
    vaux[i] := -pontos[i];
  CursorNoObjeto(vaux);
  ObjetoSelecionado(vaux, pinca, ReacaoHaptica1.Checked);
  DesenhaObjetos(true);

  //atualiza valores para prxima iterao
  tempoA := Now;
  c3d.aAnt[1] := Wiimote.Sensor.Motion.X;
  c3d.aAnt[2] := Wiimote.Sensor.Motion.Y;
  c3d.aAnt[3] := Wiimote.Sensor.Motion.Z;

  SwapBuffers(glDC);
end;

procedure TMainForm.CursorOnOff1Click(Sender: TObject);
begin
  tempoA := Now;
  c3d.tAnt := Now;
  if MotionTimer.Enabled
    then CursorOff
    else CursorOn;
end;

procedure TMainForm.AbrirCena1Click(Sender: TObject);
var i, j, f, r, nElementosOBJ: integer;
    massa: TdMass;
    q: TdQuaternion;
begin
  TimerGL.Enabled := false;

  OpenDialog1.Filter := 'Arquivo de cena (.xml)|*xml';
  OpenDialog1.InitialDir := ExtractFilePath(Application.ExeName) + 'Cenas';
  if OpenDialog1.Execute then
    begin
    AbrirCena(ClientDataSet1, OpenDialog1.FileName);

    TemGravidade := ClientDataSet1.FieldByName('TemGravidade').AsBoolean;
    ShutDownODE;
    InitODE;
    // carrega modelos
    with ClientDataSet1 do
      begin
      // limpa as variveis para usar novamente
      SetLength(objetosCena.Objetos,0);
      SetLength(Geom,0);
      SetLength(Body,0);
      SetLength(Atrito,0);

      nElementosOBJ := RecordCount;
      SetLength(objetosCena.Objetos,nElementosOBJ);
      SetLength(Geom,nElementosOBJ);
      SetLength(Body,nElementosOBJ);
      SetLength(Atrito,nElementosOBJ);
      TipoDeColisao := FieldByName('TipoDeColisao').AsInteger;
      for r := 0 to RecordCount-1 do
        begin
        RecNo := r+1;
        Body[r] := dBodyCreate(m_ODEWorld);
        Atrito[r] := FieldByName('Atrito').AsFloat;
        case FieldByName('ODETipo').AsInteger of
          0: begin
             Geom[r] := dCreateSphere (m_ODESpace,
                                       FieldByName('Dim1').AsFloat {raio});
             dGeomSphereSetRadius(Geom[r],
                                  FieldByName('Dim1').AsFloat);

             dGeomSetBody(Geom[r], Body[r]);
             dGeomSetPosition(Geom[r],
                              FieldByName('X').AsFloat,
                              FieldByName('Y').AsFloat,
                              FieldByName('Z').AsFloat);

             // *****************************
             // http://blog.gmane.org/gmane.comp.lib.ode/month=20070101/page=8
             // http://veazie.org/Images/odetest.cpp.txt
             // *****************************
             q[0] := DegToRad(FieldByName('Ang').AsFloat);
             q[1] := FieldByName('eixoX').AsFloat;
             q[2] := FieldByName('eixoY').AsFloat;
             q[3] := FieldByName('eixoZ').AsFloat;
             dGeomSetQuaternion(Geom[r],q);

             dMassSetSphereTotal(massa,
                                 FieldByName('Massa').AsFloat,
                                 FieldByName('Dim1').AsFloat);
             dBodySetMass(Body[r], @massa);

             for i := 1 to 3 do
               objetosCena.Objetos[r].eixoRot[i] := q[i];
             objetosCena.Objetos[r].angRot := q[0];

             objetosCena.Objetos[r].tipo := otEsfera;
             objetosCena.Objetos[r].size := FieldByName('Dim1').AsFloat;
             objetosCena.Objetos[r].D1 := FieldByName('Dim1').AsFloat;
             end;
          1: begin
             Geom[r] := dCreateCCylinder(m_ODESpace, FieldByName('Dim1').AsFloat {raio}, FieldByName('Dim2').AsFloat {comprimento});
             dGeomCCylinderSetParams(Geom[r], FieldByName('Dim1').AsFloat, FieldByName('Dim2').AsFloat);

             dGeomSetBody(Geom[r], Body[r]);
             dGeomSetPosition(Geom[r],
                              FieldByName('X').AsFloat,
                              FieldByName('Y').AsFloat,
                              FieldByName('Z').AsFloat);

             q[0] := DegToRad(FieldByName('Ang').AsFloat);
             q[1] := FieldByName('eixoX').AsFloat;
             q[2] := FieldByName('eixoY').AsFloat;
             q[3] := FieldByName('eixoZ').AsFloat;
             dGeomSetQuaternion(Geom[r],q);

             for i := 1 to 3 do
               objetosCena.Objetos[r].eixoRot[i] := q[i];
             objetosCena.Objetos[r].angRot := q[0];

             // dMassSetCappedCylinderTotal -> dir -> x=1, y=2, z=3
             dMassSetCappedCylinderTotal(massa,
                                         FieldByName('Massa').AsFloat,
                                         3 {z},
                                         FieldByName('Dim1').AsFloat,
                                         FieldByName('Dim2').AsFloat);
             dBodySetMass(Body[r], @massa);

             objetosCena.Objetos[r].tipo := otCapsula;
             objetosCena.Objetos[r].size := 2 * FieldByName('Dim1').AsFloat + FieldByName('Dim2').AsFloat;
             objetosCena.Objetos[r].D1 := FieldByName('Dim1').AsFloat;
             objetosCena.Objetos[r].D2 := FieldByName('Dim2').AsFloat;
             end;
          2: begin
             Geom[r] := dCreateBox(m_ODESpace,
                                   FieldByName('Dim1').AsFloat {Lx},
                                   FieldByName('Dim2').AsFloat {Ly},
                                   FieldByName('Dim2').AsFloat {Lz});
             dGeomBoxSetLengths(Geom[r],
                                FieldByName('Dim1').AsFloat,
                                FieldByName('Dim2').AsFloat,
                                FieldByName('Dim3').AsFloat);
             dGeomSetBody(Geom[r], Body[r]);
             dGeomSetPosition(Geom[r],
                              FieldByName('X').AsFloat,
                              FieldByName('Y').AsFloat,
                              FieldByName('Z').AsFloat);

             q[0] := DegToRad(FieldByName('Ang').AsFloat);
             q[1] := FieldByName('eixoX').AsFloat;
             q[2] := FieldByName('eixoY').AsFloat;
             q[3] := FieldByName('eixoZ').AsFloat;
             dGeomSetQuaternion(Geom[r],q);

             for i := 1 to 3 do
               objetosCena.Objetos[r].eixoRot[i] := q[i];
             objetosCena.Objetos[r].angRot := q[0];

             dMassSetBoxTotal(massa,
                              FieldByName('Massa').AsFloat,
                              FieldByName('Dim1').AsFloat,
                              FieldByName('Dim2').AsFloat,
                              FieldByName('Dim3').AsFloat);
             dBodySetMass(Body[r], @massa);

             objetosCena.Objetos[r].tipo := otCubo;
             objetosCena.Objetos[r].size := sqrt(
               sqr(FieldByName('Dim1').AsFloat/2)+
               sqr(FieldByName('Dim2').AsFloat/2)+
               sqr(FieldByName('Dim3').AsFloat/2));
             objetosCena.Objetos[r].D1 := FieldByName('Dim1').AsFloat;
             objetosCena.Objetos[r].D2 := FieldByName('Dim2').AsFloat;
             objetosCena.Objetos[r].D3 := FieldByName('Dim3').AsFloat;
           end;
          3: begin
             Geom[r] := dCreatePlane(m_ODESpace,
                                     FieldByName('Dim1').AsFloat {a},
                                     FieldByName('Dim2').AsFloat {b},
                                     FieldByName('Dim3').AsFloat {c},
                                     FieldByName('Dim4').AsFloat {d});
             dGeomPlaneSetParams(Geom[r],
                                 FieldByName('Dim1').AsFloat,
                                 FieldByName('Dim2').AsFloat,
                                 FieldByName('Dim3').AsFloat,
                                 FieldByName('Dim4').AsFloat);

             dGeomSetBody(Geom[r], Body[r]);

             //dBodyDisable(Body[r]);
             dBodySetGravityMode(Body[r],0);


             objetosCena.Objetos[r].Modelo.BoundingBox.med.X := FieldByName('SX').AsFloat;
             objetosCena.Objetos[r].Modelo.BoundingBox.med.Y := FieldByName('SY').AsFloat;

             objetosCena.Objetos[r].tipo := otPlano;
             objetosCena.Objetos[r].size := sqrt(
               sqr(FieldByName('SX').AsFloat/2)+
               sqr(FieldByName('SY').AsFloat/2));
             objetosCena.Objetos[r].D1 := FieldByName('Dim1').AsFloat;
             objetosCena.Objetos[r].D2 := FieldByName('Dim2').AsFloat;
             objetosCena.Objetos[r].D3 := FieldByName('Dim3').AsFloat;
             objetosCena.Objetos[r].D4 := FieldByName('Dim4').AsFloat;
             end;
          4: begin
             // limpa quebra de linha do OBJ
             LimpaOBJ(FieldByName('Arquivo').AsString);
             objetosCena.Objetos[r].Modelo :=LoadModel(FieldByName('Arquivo').AsString);
             // no usa vrtices repetidos para criar as faces
             VerticesUnicosOBJ(objetosCena.Objetos[r].Modelo);
             // mudar o tamanho de cada modelo
             EscalaOBJ(objetosCena.Objetos[r].Modelo,
                       FieldByName('SX').AsFloat,
                       FieldByName('SY').AsFloat,
                       FieldByName('SZ').AsFloat);
             // dimenses do modelo
             GetMinMax(objetosCena.Objetos[r].Modelo);

             // Leitura do modelo do arquivo
             for i := 1 to objetosCena.Objetos[r].Modelo.Vertices do
               begin
               Vertices[i-1][0] := objetosCena.Objetos[r].Modelo.Vertex[i].X;
               Vertices[i-1][1] := objetosCena.Objetos[r].Modelo.Vertex[i].Y;
               Vertices[i-1][2] := objetosCena.Objetos[r].Modelo.Vertex[i].Z;
               end;

             f := 0;
             for i := 1 to objetosCena.Objetos[r].Modelo.Groups do
               for j := 1 to objetosCena.Objetos[r].Modelo.Group[i].Faces do
                 begin
                 Indices[f] := objetosCena.Objetos[r].Modelo.Group[i].Face[j].vIndex[0];
                 Indices[f+1] := objetosCena.Objetos[r].Modelo.Group[i].Face[j].vIndex[1];
                 Indices[f+2] := objetosCena.Objetos[r].Modelo.Group[i].Face[j].vIndex[2];
                 inc(f,3);
                 end;

             vPosTriMesh.Pos := toVector3f(FieldByName('X').AsFloat,
                                           FieldByName('Y').AsFloat,
                                           FieldByName('Z').AsFloat);
             vPosTriMesh.Size.x := objetosCena.Objetos[r].Modelo.BoundingBox.max.X-objetosCena.Objetos[r].Modelo.BoundingBox.min.X;
             vPosTriMesh.Size.y := objetosCena.Objetos[r].Modelo.BoundingBox.max.Y-objetosCena.Objetos[r].Modelo.BoundingBox.min.Y;
             vPosTriMesh.Size.z := objetosCena.Objetos[r].Modelo.BoundingBox.max.Z-objetosCena.Objetos[r].Modelo.BoundingBox.min.Z;

             // Associao dos dados do modelo com a ODE
             TriMeshData := dGeomTriMeshDataCreate();
             dGeomTriMeshDataBuildSimple(TriMeshData,
                                         @Vertices, objetosCena.Objetos[r].Modelo.Vertices,
                                         @Indices, f div 3);

             // create the trimesh's collision geometry
             Geom[r] := dCreateTriMesh(m_ODESpace, TriMeshData, nil, nil, nil);
             // attach the geometry to the trimesh
             dGeomSetBody(Geom[r], Body[r]);
             // set the trimesh's position
             dBodySetPosition(Body[r],
                              vPosTriMesh.Pos.x,
                              vPosTriMesh.Pos.y ,
                              vPosTriMesh.Pos.z);

             q[0] := DegToRad(FieldByName('Ang').AsFloat);
             q[1] := FieldByName('eixoX').AsFloat;
             q[2] := FieldByName('eixoY').AsFloat;
             q[3] := FieldByName('eixoZ').AsFloat;
             dGeomSetQuaternion(Geom[r],q);

             for i := 1 to 3 do
               objetosCena.Objetos[r].eixoRot[i] := q[i];
             objetosCena.Objetos[r].angRot := q[0];

             dMassSetBoxTotal(massa,
                              FieldByName('Massa').AsFloat,
                              vPosTriMesh.Size.x,
                              vPosTriMesh.Size.y ,
                              vPosTriMesh.Size.z);
             dBodySetMass(Body[r],@massa);

             objetosCena.Objetos[r].tipo := otTriMesh;
             objetosCena.Objetos[r].size := sqrt(
               sqr(objetosCena.Objetos[r].Modelo.BoundingBox.max.X-objetosCena.Objetos[r].Modelo.BoundingBox.med.X)+
               sqr(objetosCena.Objetos[r].Modelo.BoundingBox.max.Y-objetosCena.Objetos[r].Modelo.BoundingBox.med.Y)+
               sqr(objetosCena.Objetos[r].Modelo.BoundingBox.max.Z-objetosCena.Objetos[r].Modelo.BoundingBox.med.Z));
             end;
          end; // case

        objetosCena.Objetos[r].p[1] := dBodyGetPosition(Body[r])[0];
        objetosCena.Objetos[r].p[2] := dBodyGetPosition(Body[r])[1];
        objetosCena.Objetos[r].p[3] := dBodyGetPosition(Body[r])[2];
        objetosCena.Objetos[r].pant[1] := dBodyGetPosition(Body[r])[0];
        objetosCena.Objetos[r].pant[2] := dBodyGetPosition(Body[r])[1];
        objetosCena.Objetos[r].pant[3] := dBodyGetPosition(Body[r])[2];
        objetosCena.Objetos[r].noCursor := false;
        objetosCena.Objetos[r].selecionado := false;
        objetosCena.Objetos[r].cor.R := FieldByName('R').AsFloat;
        objetosCena.Objetos[r].cor.G := FieldByName('G').AsFloat;
        objetosCena.Objetos[r].cor.B := FieldByName('B').AsFloat;
        objetosCena.Objetos[r].cor.A := FieldByName('A').AsFloat;
        objetosCena.Objetos[r].MostraRastroSempre := FieldByName('MostraRastroSempre').AsBoolean;
        objetosCena.Objetos[r].MostraRastroSolto := FieldByName('MostraRastroSolto').AsBoolean;
        objetosCena.Objetos[r].MostraVecV := FieldByName('MostraVecV').AsBoolean;
        objetosCena.Objetos[r].MostraVecA := FieldByName('MostraVecA').AsBoolean;
        objetosCena.Objetos[r].ODEMove := FieldByName('ODEMove').AsBoolean;
        objetosCena.Objetos[r].UserMove := FieldByName('UserMove').AsBoolean;
        objetosCena.Objetos[r].MoveDir[1] := FieldByName('MoveX').AsBoolean;
        objetosCena.Objetos[r].MoveDir[2] := FieldByName('MoveY').AsBoolean;
        objetosCena.Objetos[r].MoveDir[3] := FieldByName('MoveZ').AsBoolean;
        SetLength(objetosCena.Objetos[r].rastro, 0);
        if not objetosCena.Objetos[r].ODEMove then
          dBodyDisable(Body[r]);

        end; // for r
      end; // with CDS
    end; // if OpenDialog
    TimerGL.Enabled := true;
end;

procedure TMainForm.ReacaoHaptica1Click(Sender: TObject);
begin
  ReacaoHaptica1.Checked := not ReacaoHaptica1.Checked;
end;

end.

