unit unitWiinRemote;

interface

uses Windows,Classes,Graphics,Forms,SysUtils;

function AppPath(): String;
function Replace(Str, Before, After: string): string;
function IsDisplayEdge(curPoint:TPoint):boolean;
function IfThen(AValue: Boolean; const ATrue: string; const AFalse: string): string;overload;
function IfThen(AValue: Boolean; const ATrue: TColor; const AFalse: TColor): TColor;overload;
function Roundoff(X: Extended): Longint;
procedure LoadConfig();
procedure SaveConfig();
function IsEventRepeat(AssignType:string):boolean;
function IsCursorEvent(AssignType:string):boolean;

type

  TAxisData = record
    X: Byte;
    Y: Byte;
    Z: Byte;
  end;

  TIntAxisData = record
    X: Integer;
    Y: Integer;
    Z: Integer;
  end;

  TButtonAssign = record
    AssignType:string;
    AssignValue:string;
  end;

  TWiimoteButtonAssign = record
    Up:TButtonAssign;
    Down:TButtonAssign;
    Right:TButtonAssign;
    Left:TButtonAssign;
    ButtonA:TButtonAssign;
    ButtonB:TButtonAssign;
    Home:TButtonAssign;
    Plus:TButtonAssign;
    Minus:TButtonAssign;
    Button1:TButtonAssign;
    Button2:TButtonAssign;

    ButtonC:TButtonAssign;
    ButtonZ:TButtonAssign;
  end;

  TButtonState = record
    PushFlag:boolean;
    LastPushTime:Cardinal;
  end;

  TWiimoteCalibration = record
    MotionSensor:TAxisData;
    NunchukMotionSensor:TAxisData;
  end;

  TWiinRemoteSetting = record
    CursorMode:string;
    MotionSensorThreshold:integer;
    InvertH:boolean;
    InvertV:boolean;
    NunchukAnalogInvertH:boolean;
    NunchukAnalogInvertV:boolean;
    CursorSpeed:string;
    Calibration:TWiimoteCalibration;
    RumbleOnEdge:boolean;
    UseMotionWhenLostIR:boolean;
    DrawGraph:boolean;
    MotionSpeedFast:integer;
    MotionSpeedSlow:integer;
    AnalogCursorSpeed:integer;
    MinimizeAtStartup:boolean;
    isNunchukEnabled:boolean;
    ReCalibrate:boolean;
    EnableCursorAtStartup:boolean;
    DisplayEdgeAdjustment:integer;
  end;

  TWiimoteIR = record
    dot1X:integer;
    dot1Y:integer;
    dot1S:integer;
    dot2X:integer;
    dot2Y:integer;
    dot2S:integer;

    LeftIs1:boolean;
    UpperIs1:boolean;

    ScreenX:integer;
    ScreenY:integer;

    DX:integer;
    DY:integer;
  end;

  TWiimoteButtonState = record
    Up:TButtonState;
    Down:TButtonState;
    Right:TButtonState;
    Left:TButtonState;
    ButtonA:TButtonState;
    ButtonB:TButtonState;
    Home:TButtonState;
    Plus:TButtonState;
    Minus:TButtonState;
    Button1:TButtonState;
    Button2:TButtonState;

    ButtonC:TButtonState;
    ButtonZ:TButtonState;
  end;

  TNunchuk = record
    Analog:TIntAxisData;
    Motion:TIntAxisData;
  end;

  TWiimoteSensorState = record
    IR:TWiimoteIR;
    Motion:TAxisData;
    Nunchuk:TNunchuk;
  end;


  TWiimote = record
    Assign:TWiimoteButtonAssign;
    Button:TWiimoteButtonState;
    Sensor:TWiimoteSensorState;
    Battery:integer;
    isNunchukEnabled:boolean;
    isClassicControllerEnabled:boolean;
  end;

const
  WII_REMOTE_VID = $057E;
  WII_REMOTE_PID = $0306;

  WII_RID_LEDS     = $11;
  WII_RID_MODE     = $12;
  WII_RID_IR_EN    = $13;
  WII_RID_SPK_EN   = $14;
  WII_RID_STATUS   = $15;
  WII_RID_WMEM     = $16;
  WII_RID_RMEM     = $17;
  WII_RID_SPK      = $18;
  WII_RID_SPK_MUTE = $19;
  WII_RID_IR_EN2   = $1A;

  WII_MODE_EXPANSION  = $20;
  WII_MODE_BASIC      = $30;
  WII_MODE_ACC        = $31;
  WII_MODE_ACC_IR     = $33;
  WII_MODE_ACC_IR_EXT = $37;
  WII_MODE_FULL       = $3E;

{
Button                                                  Number (dec)   Value (hex)
Two                                                         1             0x0001
One                                                         2             0x0002
B                                                           3             0x0004
A                                                           4             0x0008
Minus                                                       5             0x0010
Z Acceleration, bit 6 (report 3E) or bit 2 (report 3F)      6             0x0020
Z Acceleration, bit 7 (report 3E) or bit 3 (report 3F)      7             0x0040
Home                                                        8             0x0080
Left                                                        9             0x0100
Right                                                      10             0x0200
Down                                                       11             0x0400
Up                                                         12             0x0800
Plus                                                       13             0x1000
Z Acceleration, bit 4 (report 3E) or bit 0 (report 3F)     14             0x2000
Z Acceleration, bit 5 (report 3E) or bit 1 (report 3F)     15             0x4000
? Unknown ?                                                16             0x8000
}

	WII_BUTTON_C		 = $4000;
	WII_BUTTON_Z		 = $2000;

  WII_BUTTON_PLUS  = $1000;
  WII_BUTTON_UP    = $0800;
  WII_BUTTON_DOWN  = $0400;
  WII_BUTTON_RIGHT = $0200;
  WII_BUTTON_LEFT  = $0100;

  WII_BUTTON_HOME  = $0080;
  WII_BUTTON_MINUS = $0010;

  WII_BUTTON_A     = $0008;
  WII_BUTTON_B     = $0004;
  WII_BUTTON_1     = $0002;
  WII_BUTTON_2     = $0001;

  WII_EVENT_DOWN   = 1;
  WII_EVENT_UP     = 2;

  GRAPH_DIV=4;
  AXIS_DIV =4;
  IR_DIV =8;

  WHEEL_FORWARD = 120 ;
  WHEEL_BACKWARD = -120;

  STR_BUTTON_ASSIGN_TYPE_KEY              = 'Keys';
  STR_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF     = 'Cursor ON/OFF';
  STR_BUTTON_ASSIGN_TYPE_CURSOR_ON        = 'Cursor ON';

  STR_BUTTON_ASSIGN_TYPE_CURSOR_UP        = 'Mouse->Cursor Up';
  STR_BUTTON_ASSIGN_TYPE_CURSOR_DOWN      = 'Mouse->Cursor Down';
  STR_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT     = 'Mouse->Cursor Right';
  STR_BUTTON_ASSIGN_TYPE_CURSOR_LEFT      = 'Mouse->Cursor Left';

  STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT       = 'Mouse->Left Click';
  STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL   = 'Mouse->Left Double Click';
  STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT      = 'Mouse->Right Click';
  STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL  = 'Mouse->Right Double Click';
  STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE     = 'Mouse->Middle Click';
  STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL = 'Mouse->Middle Double Click';

  STR_BUTTON_ASSIGN_TYPE_WHEEL_UP         = 'Mouse->Wheel Up';
  STR_BUTTON_ASSIGN_TYPE_WHEEL_DOWN       = 'Mouse->Wheel Down';
  STR_BUTTON_ASSIGN_TYPE_WEB_FORWARD      = 'Web->Forward';
  STR_BUTTON_ASSIGN_TYPE_WEB_BACKWARD     = 'Web->Backward';
  STR_BUTTON_ASSIGN_TYPE_VOLUME_UP        = 'Mixer->Volume Up';
  STR_BUTTON_ASSIGN_TYPE_VOLUME_DOWN      = 'Mixer->Volume Down';
  STR_BUTTON_ASSIGN_TYPE_VOLUME_MUTE      = 'Mixer->Mute';
  STR_BUTTON_ASSIGN_TYPE_EXECUTE          = 'Execute';

  INT_BUTTON_ASSIGN_TYPE_KEY              = 0;
  INT_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF     = 1;
  INT_BUTTON_ASSIGN_TYPE_CURSOR_ON        = 2;

  INT_BUTTON_ASSIGN_TYPE_CURSOR_UP        = 3;
  INT_BUTTON_ASSIGN_TYPE_CURSOR_DOWN      = 4;
  INT_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT     = 5;
  INT_BUTTON_ASSIGN_TYPE_CURSOR_LEFT      = 6;

  INT_BUTTON_ASSIGN_TYPE_MOUSE_LEFT       = 7;
  INT_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL   = 8;
  INT_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT      = 9;
  INT_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL  = 10;
  INT_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE     = 11;
  INT_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL = 12;
  INT_BUTTON_ASSIGN_TYPE_WHEEL_UP         = 13;
  INT_BUTTON_ASSIGN_TYPE_WHEEL_DOWN       = 14;
  INT_BUTTON_ASSIGN_TYPE_WEB_FORWARD      = 15;
  INT_BUTTON_ASSIGN_TYPE_WEB_BACKWARD     = 16;
  INT_BUTTON_ASSIGN_TYPE_VOLUME_UP        = 17;
  INT_BUTTON_ASSIGN_TYPE_VOLUME_DOWN      = 18;
  INT_BUTTON_ASSIGN_TYPE_VOLUME_MUTE      = 19;
  INT_BUTTON_ASSIGN_TYPE_EXECUTE          = 20;

  //for repeat
  BOOL_BUTTON_ASSIGN_TYPE_KEY              = true;
  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF     = false;
  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_ON        = false;

  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_UP        = false;
  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_DOWN      = false;
  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT     = false;
  BOOL_BUTTON_ASSIGN_TYPE_CURSOR_LEFT      = false;

  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_LEFT       = false;
  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL   = false;
  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT      = false;
  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL  = false;
  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE     = false;
  BOOL_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL = false;
  BOOL_BUTTON_ASSIGN_TYPE_WHEEL_UP         = true;
  BOOL_BUTTON_ASSIGN_TYPE_WHEEL_DOWN       = true;
  BOOL_BUTTON_ASSIGN_TYPE_WEB_FORWARD      = true;
  BOOL_BUTTON_ASSIGN_TYPE_WEB_BACKWARD     = true;
  BOOL_BUTTON_ASSIGN_TYPE_VOLUME_UP        = true;
  BOOL_BUTTON_ASSIGN_TYPE_VOLUME_DOWN      = true;
  BOOL_BUTTON_ASSIGN_TYPE_VOLUME_MUTE      = false;
  BOOL_BUTTON_ASSIGN_TYPE_EXECUTE          = false;

  INT_BUTTON_ASSIGN_UP    = 0;
  INT_BUTTON_ASSIGN_DOWN  = 1;
  INT_BUTTON_ASSIGN_RIGHT = 2;
  INT_BUTTON_ASSIGN_LEFT  = 3;
  INT_BUTTON_ASSIGN_A     = 4;
  INT_BUTTON_ASSIGN_B     = 5;
  INT_BUTTON_ASSIGN_HOME  = 6;
  INT_BUTTON_ASSIGN_PLUS  = 7;
  INT_BUTTON_ASSIGN_MINUS = 8;
  INT_BUTTON_ASSIGN_1     = 9;
  INT_BUTTON_ASSIGN_2     = 10;

  INT_BUTTON_ASSIGN_C     = 11;
  INT_BUTTON_ASSIGN_Z     = 12;

  STR_CURSOR_MODE_IR ='IR';
  STR_CURSOR_MODE_MOTION ='Motion';

  STR_CURSOR_MODE_NUNCHUK_MOTION ='Nunchuk Motion';
  STR_CURSOR_MODE_NUNCHUK_ANALOG ='Nunchuk Analog';

  STR_CURSOR_SPEED_FASTER ='Faster';
  STR_CURSOR_SPEED_SLOWER ='Slower';

var

    Wiimote:TWiimote;

    Setting:TWiinRemoteSetting;

    //Mixer
    DestinationIndex,ConnectionIndex:integer;


implementation
uses IniFiles;
procedure LoadConfig();
var
  INI:TIniFile;
begin
  INI:=TIniFile.Create(AppPath+'config.ini');

  with Setting do
  begin
    CursorMode:=INI.ReadString('General','CursorMode',STR_CURSOR_MODE_MOTION);

    EnableCursorAtStartup:=INI.ReadBool('General','EnableCursorAtStartup',false);

    RumbleOnEdge:=INI.ReadBool('General','RumbleOnEdge',true);

    UseMotionWhenLostIR:=INI.ReadBool('General','UseMotionWhenLostIR',false);

    MinimizeAtStartup:=INI.ReadBool('General','MinimizeAtStartup',false);
    DrawGraph:=INI.ReadBool('General','DrawGraph',true);
    DisplayEdgeAdjustment:=INI.ReadInteger('General','DisplayEdgeAdjustment',50);

    MotionSensorThreshold:=INI.ReadInteger('Motion','SensorThreshold',2);
    CursorSpeed:=INI.ReadString('Motion','OnPush',STR_CURSOR_SPEED_SLOWER);

    InvertH:=INI.ReadBool('Motion','InvertH',false);
    InvertV:=INI.ReadBool('Motion','InvertV',false);

    NunchukAnalogInvertH:=INI.ReadBool('Nunchuk','AnalogInvertH',false);
    NunchukAnalogInvertV:=INI.ReadBool('Nunchuk','AnalogInvertV',false);

    Calibration.MotionSensor.X:=INI.ReadInteger('Motion','BaseX',128);
    Calibration.MotionSensor.Y:=INI.ReadInteger('Motion','BaseY',128);
    Calibration.MotionSensor.Z:=INI.ReadInteger('Motion','BaseZ',160);


    MotionSpeedFast:=INI.ReadInteger('Motion','SpeedFast',5);
    MotionSpeedSlow:=INI.ReadInteger('Motion','SpeedSlow',3);

    AnalogCursorSpeed:=INI.ReadInteger('Nunchuk','AnalogCursorSpeed',8);

    ReCalibrate:=INI.ReadBool('Motion','Recalibrate',false);

   Setting.isNunchukEnabled:=((CursorMode=STR_CURSOR_MODE_NUNCHUK_MOTION) or (CursorMode=STR_CURSOR_MODE_NUNCHUK_ANALOG));

  end;

  with Wiimote.Assign do
  begin
    Up.AssignType:=INI.ReadString('Up','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Up.AssignValue:=INI.ReadString('Up','Value','Up');

    Down.AssignType:=INI.ReadString('Down','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Down.AssignValue:=INI.ReadString('Down','Value','Down');

    Right.AssignType:=INI.ReadString('Right','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Right.AssignValue:=INI.ReadString('Right','Value','Right');

    Left.AssignType:=INI.ReadString('Left','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Left.AssignValue:=INI.ReadString('Left','Value','Left');

    ButtonA.AssignType:=INI.ReadString('A','Type',STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT);
    ButtonA.AssignValue:=INI.ReadString('A','Value','');

    ButtonB.AssignType:=INI.ReadString('B','Type',STR_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF);
    ButtonB.AssignValue:=INI.ReadString('B','Value','');

    Home.AssignType:=INI.ReadString('Home','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Home.AssignValue:=INI.ReadString('Home','Value','BkSp');

    Plus.AssignType:=INI.ReadString('Plus','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Plus.AssignValue:=INI.ReadString('Plus','Value','Enter');

    Minus.AssignType:=INI.ReadString('Minus','Type',STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT);
    Minus.AssignValue:=INI.ReadString('Minus','Value','');

    Button1.AssignType:=INI.ReadString('1','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Button1.AssignValue:=INI.ReadString('1','Value','Tab');

    Button2.AssignType:=INI.ReadString('2','Type',STR_BUTTON_ASSIGN_TYPE_KEY);
    Button2.AssignValue:=INI.ReadString('2','Value','Enter');

    ButtonC.AssignType:=INI.ReadString('C','Type',STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT);
    ButtonC.AssignValue:=INI.ReadString('C','Value','');

    ButtonZ.AssignType:=INI.ReadString('Z','Type',STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT);
    ButtonZ.AssignValue:=INI.ReadString('Z','Value','');

  end;
  INI.Free;

end;

procedure SaveConfig();
var
  INI:TIniFile;
begin

  with Setting do
  begin
    INI:=TIniFile.Create(AppPath+'config.ini');

    INI.WriteString('General','CursorMode',CursorMode);

    INI.WriteBool('General','RumbleOnEdge',RumbleOnEdge);
    INI.WriteBool('General','UseMotionWhenLostIR',UseMotionWhenLostIR);
    INI.WriteBool('General','MinimizeAtStartup',MinimizeAtStartup);
    INI.WriteBool('General','EnableCursorAtStartup',EnableCursorAtStartup);
    INI.WriteBool('General','DrawGraph',DrawGraph);

    INI.WriteInteger('General','DisplayEdgeAdjustment',DisplayEdgeAdjustment);


    INI.WriteInteger('Motion','SensorThreshold',MotionSensorThreshold);
    INI.WriteString('Motion','OnPush',CursorSpeed);


    INI.WriteBool('Motion','InvertH',InvertH);
    INI.WriteBool('Motion','InvertV',InvertV);

    INI.WriteBool('Nunchuk','AnalogInvertH',NunchukAnalogInvertH);
    INI.WriteBool('Nunchuk','AnalogInvertV',NunchukAnalogInvertV);

    INI.WriteInteger('Motion','BaseX',Calibration.MotionSensor.X);
    INI.WriteInteger('Motion','BaseY',Calibration.MotionSensor.Y);
    INI.WriteInteger('Motion','BaseZ',Calibration.MotionSensor.Z);

    INI.WriteInteger('Motion','SpeedFast',MotionSpeedFast);
    INI.WriteInteger('Motion','SpeedSlow',MotionSpeedSlow);

    INI.WriteInteger('Nunchuk','AnalogCursorSpeed',AnalogCursorSpeed);
    INI.WriteBool('Motion','Recalibrate',ReCalibrate);
  end;

  with Wiimote.Assign do
  begin
    INI.WriteString('Up','Type',Up.AssignType);
    INI.WriteString('Up','Value',Up.AssignValue);

    INI.WriteString('Down','Type',Down.AssignType);
    INI.WriteString('Down','Value',Down.AssignValue);

    INI.WriteString('Right','Type',Right.AssignType);
    INI.WriteString('Right','Value',Right.AssignValue);

    INI.WriteString('Left','Type',Left.AssignType);
    INI.WriteString('Left','Value',Left.AssignValue);


    INI.WriteString('A','Type',ButtonA.AssignType);
    INI.WriteString('A','Value',ButtonA.AssignValue);

    INI.WriteString('B','Type',ButtonB.AssignType);
    INI.WriteString('B','Value',ButtonB.AssignValue);

    INI.WriteString('Home','Type',Home.AssignType);
    INI.WriteString('Home','Value',Home.AssignValue);

    INI.WriteString('Plus','Type',Plus.AssignType);
    INI.WriteString('Plus','Value',Plus.AssignValue);

    INI.WriteString('Minus','Type',Minus.AssignType);
    INI.WriteString('Minus','Value',Minus.AssignValue);

    INI.WriteString('1','Type',Button1.AssignType);
    INI.WriteString('1','Value',Button1.AssignValue);

    INI.WriteString('2','Type',Button2.AssignType);
    INI.WriteString('2','Value',Button2.AssignValue);

    INI.WriteString('C','Type',ButtonC.AssignType);
    INI.WriteString('C','Value',ButtonC.AssignValue);

    INI.WriteString('Z','Type',ButtonZ.AssignType);
    INI.WriteString('Z','Value',ButtonZ.AssignValue);
  end;
  INI.UpdateFile;

  INI.Free;
end;
function IsEventRepeat(AssignType:string):boolean;
begin
  Result:=false;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_KEY              then Result:=BOOL_BUTTON_ASSIGN_TYPE_KEY;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF     then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_ONOFF;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_ON        then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_ON;

  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_UP        then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_UP;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_DOWN      then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_DOWN;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT     then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_LEFT      then Result:=BOOL_BUTTON_ASSIGN_TYPE_CURSOR_LEFT;

  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT       then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_LEFT;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL   then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_LEFT_DBL;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT      then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL  then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_RIGHT_DBL;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE     then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL then Result:=BOOL_BUTTON_ASSIGN_TYPE_MOUSE_MIDDLE_DBL;

  if AssignType =STR_BUTTON_ASSIGN_TYPE_WHEEL_UP         then Result:=BOOL_BUTTON_ASSIGN_TYPE_WHEEL_UP;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_WHEEL_DOWN       then Result:=BOOL_BUTTON_ASSIGN_TYPE_WHEEL_DOWN;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_WEB_FORWARD      then Result:=BOOL_BUTTON_ASSIGN_TYPE_WEB_FORWARD;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_WEB_BACKWARD     then Result:=BOOL_BUTTON_ASSIGN_TYPE_WEB_BACKWARD;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_VOLUME_UP        then Result:=BOOL_BUTTON_ASSIGN_TYPE_VOLUME_UP;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_VOLUME_DOWN      then Result:=BOOL_BUTTON_ASSIGN_TYPE_VOLUME_DOWN;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_VOLUME_MUTE      then Result:=BOOL_BUTTON_ASSIGN_TYPE_VOLUME_MUTE;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_EXECUTE          then Result:=BOOL_BUTTON_ASSIGN_TYPE_EXECUTE;
end;

function IsCursorEvent(AssignType:string):boolean;
var
  ret:boolean;
begin
  ret:=false;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_UP        then ret:=true;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_DOWN      then ret:=true;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_RIGHT     then ret:=true;
  if AssignType =STR_BUTTON_ASSIGN_TYPE_CURSOR_LEFT      then ret:=true;

  result:=ret;
end;

function IsDisplayEdge(curPoint:TPoint):boolean;
var
  ret1,ret2,ret3,ret4:boolean;
begin
  ret1:=(Screen.MonitorFromPoint(Point(curPoint.X-1,curPoint.Y),mdNull)=nil);
  ret2:=(Screen.MonitorFromPoint(Point(curPoint.X+1,curPoint.Y),mdNull)=nil);
  ret3:=(Screen.MonitorFromPoint(Point(curPoint.X,curPoint.Y+1),mdNull)=nil);
  ret4:=(Screen.MonitorFromPoint(Point(curPoint.X,curPoint.Y-1),mdNull)=nil);

  result:= (ret1 OR ret2 OR ret3 OR ret4);
end;

function IfThen(AValue: Boolean; const ATrue: string; const AFalse: string): string;
begin
  if AValue then
    Result := ATrue
  else
    Result := AFalse;
end;

function IfThen(AValue: Boolean; const ATrue: TColor; const AFalse: TColor): TColor;
begin
  if AValue then
    Result := ATrue
  else
    Result := AFalse;
end;


function Roundoff(X: Extended): Longint;

begin
  if x >= 0 then Result := Trunc(x + 0.5)
  else Result := Trunc(x - 0.5);
end;

function Replace(Str, Before, After: string): string;
var
  Len, N: Integer;
  ret:string;
begin
  Len := Length(Before) - 1;
  N := AnsiPos(Before, Str);
  while N > 0 do begin
    ret := ret + Copy(Str, 1, N - 1) + After;
    Delete(Str, 1, N + Len);
    N := AnsiPos(Before, Str)
  end;
  Result := ret + Str
end;

function AppPath(): String;
begin
  Result := ExtractFilePath(Application.ExeName);
end;

end.
